from pathlib import Path
import logging
from typing import Union

import numpy as np

from freemocap.core_processes.capture_volume_calibration.triangulate_3d_data import save_mediapipe_3d_data_to_npy
from freemocap.utilities.project_3d_data_to_z_plane import project_3d_data_to_z_plane

logger = logging.getLogger(__name__)


def process_single_camera_skeleton_data(input_image_data_frame_marker_xyz: np.ndarray,
                                        raw_data_folder_path: Union[str, Path],
                                        project_to_z_plane: bool = True) -> [np.ndarray, np.ndarray]:

    if project_to_z_plane:

        logger.info("Single camera detected - "
                    "Altering image-derived 3d data to resemble multi-camera reconstructed data.")

        skeleton_reprojection_error_fr_mar = np.zeros(input_image_data_frame_marker_xyz.shape[0:2])

        raw_skel3d_frame_marker_xyz = project_3d_data_to_z_plane(
            skel3d_frame_marker_xyz=input_image_data_frame_marker_xyz)

        save_mediapipe_3d_data_to_npy(
            data3d_numFrames_numTrackedPoints_XYZ=raw_skel3d_frame_marker_xyz,
            data3d_numFrames_numTrackedPoints_reprojectionError=skeleton_reprojection_error_fr_mar,
            path_to_folder_where_data_will_be_saved=raw_data_folder_path,
        )

    return raw_skel3d_frame_marker_xyz, skeleton_reprojection_error_fr_mar



import logging
import os
import threading
from typing import Callable

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QDoubleValidator
from PyQt6.QtWidgets import (
    QFileDialog,
    QFormLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QRadioButton,
    QVBoxLayout,
    QWidget,
    QHBoxLayout,
    QLayout, )

from freemocap.core_processes.capture_volume_calibration.charuco_stuff.default_charuco_square_size import (
    default_charuco_square_size_mm,
)
from freemocap.gui.qt.workers.anipose_calibration_thread_worker import (
    AniposeCalibrationThreadWorker,
)
from freemocap.system.paths_and_files_names import (
    get_last_successful_calibration_toml_path,
)

logger = logging.getLogger(__name__)


class CalibrationControlPanel(QWidget):
    def __init__(self, get_active_recording_info_callable: Callable, kill_thread_event: threading.Event, parent=None):

        super().__init__(parent=parent)
        self._has_a_toml_path = False
        self._calibration_toml_path = None
        self._kill_thread_event = kill_thread_event

        self._get_active_recording_info_callable = get_active_recording_info_callable
        self.parent = parent

        self._anipose_calibration_frame_worker = None
        self._user_selected_calibration_toml_path_str = None

        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self._selected_calibration_toml_label = QLabel("")
        self._selected_calibration_toml_label.setWordWrap(True)

        self._layout.addWidget(self._selected_calibration_toml_label)

        self._radio_button_layout = self._create_radio_button_layout()
        self._radio_button_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self._layout.addLayout(self._radio_button_layout)
        self._layout.addStretch()

    @property
    def calibration_toml_path(self) -> str:
        return self._calibration_toml_path

    @property
    def radio_button_layout(self) -> QLayout:
        return self._radio_button_layout

    def _create_radio_button_layout(self):
        radio_button_form_layout = QFormLayout()

        radio_button_form_layout.addRow(self._create_use_most_recent_calibration_radio_button())

        radio_button_form_layout.addRow(self._create_load_calibration_from_file_radio_button())

        radio_button_form_layout.addRow(self._add_calibrate_from_active_recording_radio_button())

        self.update_calibration_toml_path()
        return radio_button_form_layout

    def _create_load_calibration_from_file_radio_button(self) -> QLayout:
        vbox_layout = QVBoxLayout()
        radio_button_push_button_layout = QHBoxLayout()
        vbox_layout.addLayout(radio_button_push_button_layout)

        self._load_calibration_from_file_radio_button = QRadioButton("Load calibration from file")
        radio_button_push_button_layout.addWidget(self._load_calibration_from_file_radio_button)
        self._load_calibration_from_file_radio_button.toggled.connect(self._handle_load_calibration_from_file_toggled)

        self._load_calibration_toml_dialog_button = QPushButton("Load Camera Calibration TOML...")
        radio_button_push_button_layout.addWidget(self._load_calibration_toml_dialog_button)
        # self._load_calibration_toml_dialog_button.setStyleSheet("font-size: 10pt;")
        self._load_calibration_toml_dialog_button.clicked.connect(self.open_load_camera_calibration_toml_dialog)
        self._load_calibration_toml_dialog_button.setEnabled(False)

        return vbox_layout

    def update_calibration_toml_path(self, toml_path: str = None):
        if toml_path is None:
            if self._load_calibration_from_file_radio_button.isChecked():
                if self._user_selected_calibration_toml_path_str is not None:
                    toml_path = self._user_selected_calibration_toml_path_str
            elif self._use_most_recent_calibration_radio_button.isChecked():
                toml_path = get_last_successful_calibration_toml_path()
            elif self._calibrate_from_active_recording_radio_button.isChecked():
                active_recording_info = self._get_active_recording_info_callable()
                if active_recording_info.calibration_toml_check:
                    toml_path = active_recording_info.calibration_toml_path
            else:
                active_rec = self._get_active_recording_info_callable()
                if active_rec is not None:
                    if active_rec.calibration_toml_check:
                        toml_path = self._get_active_recording_info_callable().calibration_toml_path
                        self._load_calibration_from_file_radio_button.setChecked(True)
                    else:
                        toml_path = get_last_successful_calibration_toml_path()
                        self._use_most_recent_calibration_radio_button.setChecked(True)



        self._calibration_toml_path = toml_path

        if self._calibration_toml_path is not None:
            logger.info(f"Setting calibration TOML path: {self._calibration_toml_path}")
            self._show_selected_calibration_toml_path(self._calibration_toml_path)
        else:
            self._show_selected_calibration_toml_path('-Calibration File Not Found-')

    def _add_calibrate_from_active_recording_radio_button(self):
        vbox = QVBoxLayout()
        hbox = QHBoxLayout()
        vbox.addLayout(hbox)

        self._calibrate_from_active_recording_radio_button = QRadioButton("Calibrate from Active Recording")
        hbox.addWidget(self._calibrate_from_active_recording_radio_button)

        self._calibrate_from_active_recording_radio_button.toggled.connect(
            self._handle_calibrate_from_active_recording_toggled
        )

        self._calibrate_from_active_recording_button = QPushButton("Calibrate from Videos")
        hbox.addWidget(self._calibrate_from_active_recording_button)

        self._calibrate_from_active_recording_button.setEnabled(False)
        self._calibrate_from_active_recording_button.clicked.connect(self.calibrate_from_active_recording)

        self._charuco_square_size_form_layout = self._create_charuco_square_size_form_layout()
        hbox2 = QHBoxLayout()
        hbox2.addStretch()
        hbox2.addLayout(self._charuco_square_size_form_layout)
        vbox.addLayout(hbox2)
        self._set_charuco_square_size_form_layout_visibility(False)

        return vbox

    def _create_use_most_recent_calibration_radio_button(self):

        self._use_most_recent_calibration_radio_button = QRadioButton(f"Use most recent calibration")

        self._use_most_recent_calibration_radio_button.toggled.connect(self._handle_use_most_recent_calibration_toggled)

        self._use_most_recent_calibration_radio_button.setToolTip(get_last_successful_calibration_toml_path())

        return self._use_most_recent_calibration_radio_button

    def update_calibrate_from_active_recording_button_text(self):
        active_recording_info = self._get_active_recording_info_callable()
        if active_recording_info is None:
            active_path_str = f"- No active recording selected -"
        else:
            if not active_recording_info.synchronized_videos_status_check:
                active_path_str = f"Recording: {active_recording_info.name} has no synchronized videos!"
            else:
                active_path_str = f"Calibrate from Recording: {self._get_active_recording_info_callable().name}"

        self._calibrate_from_active_recording_button.setToolTip(active_path_str)
        self.update_calibration_toml_path()

    def _handle_use_most_recent_calibration_toggled(self, checked):
        pass
        # if checked:
        #     self._most_recent_calibration_path_label.show()
        # else:
        #     self._most_recent_calibration_path_label.hide()
        self.update_calibration_toml_path()

    def _handle_load_calibration_from_file_toggled(self, checked):
        if checked:
            self._load_calibration_toml_dialog_button.setEnabled(True)
        else:
            self._load_calibration_toml_dialog_button.setEnabled(False)

        self.update_calibration_toml_path()

    def _handle_calibrate_from_active_recording_toggled(self, checked):
        active_recording_info = self._get_active_recording_info_callable()
        self.update_calibrate_from_active_recording_button_text()
        self.update_calibration_toml_path()

        if checked and active_recording_info is not None and active_recording_info.synchronized_videos_status_check:
            self._calibrate_from_active_recording_button.setEnabled(True)
            self._set_charuco_square_size_form_layout_visibility(True)
        else:
            self._calibrate_from_active_recording_button.setEnabled(False)
            self._set_charuco_square_size_form_layout_visibility(False)

    def _set_charuco_square_size_form_layout_visibility(self, visible):
        label_index = self._charuco_square_size_form_layout.indexOf(self._charuco_square_size_label)
        line_edit_index = self._charuco_square_size_form_layout.indexOf(self._charuco_square_size_line_edit)
        if visible:
            self._charuco_square_size_form_layout.itemAt(label_index).widget().setEnabled(True)
            self._charuco_square_size_form_layout.itemAt(line_edit_index).widget().setEnabled(True)
        else:
            self._charuco_square_size_form_layout.itemAt(label_index).widget().setEnabled(False)
            self._charuco_square_size_form_layout.itemAt(line_edit_index).widget().setEnabled(False)

    def open_load_camera_calibration_toml_dialog(self) -> str:
        # from this tutorial - https://www.youtube.com/watch?v=gg5TepTc2Jg&t=649s
        dialog = QFileDialog()
        dialog.setFileMode(QFileDialog.FileMode.ExistingFile)

        calibration_toml_path_selection = dialog.getOpenFileName(
            self,
            "Select 'toml' containing camera calibration info",
            str(self._get_active_recording_info_callable().path),
            "Camera Calibration TOML (*.toml)",
        )
        if len(calibration_toml_path_selection) > 0:
            self._user_selected_calibration_toml_path_str = calibration_toml_path_selection[0]
            logger.info(f"User selected camera calibration toml path:{self._user_selected_calibration_toml_path_str}")
            self._show_selected_calibration_toml_path(self._user_selected_calibration_toml_path_str)
            self.update_calibration_toml_path(toml_path=self._user_selected_calibration_toml_path_str)
            return self.calibration_toml_path


    def _show_selected_calibration_toml_path(self, calibration_toml_path_str: str):
        self._calibration_toml_path = calibration_toml_path_str
        path = calibration_toml_path_str.replace(os.sep, "/ ")
        self._selected_calibration_toml_label.setText(path)
        self._selected_calibration_toml_label.show()

    def _create_charuco_square_size_form_layout(self):
        charuco_square_size_form_layout = QFormLayout()

        self._charuco_square_size_line_edit = QLineEdit()
        self._charuco_square_size_line_edit.setValidator(QDoubleValidator())
        self._charuco_square_size_line_edit.setFixedWidth(100)

        self._charuco_square_size_label = QLabel("Charuco square size (mm)")
        self._charuco_square_size_label.setStyleSheet("QLabel { font-size: 12px;  }")

        self._charuco_square_size_line_edit.setText(str(default_charuco_square_size_mm))
        self._charuco_square_size_line_edit.setToolTip(
            "The length of one of the edges of the black squares in the calibration board in mm"
        )
        charuco_square_size_form_layout.addRow(self._charuco_square_size_label, self._charuco_square_size_line_edit)
        charuco_square_size_form_layout.setAlignment(Qt.AlignmentFlag.AlignRight)
        return charuco_square_size_form_layout

    def calibrate_from_active_recording(self, charuco_square_size_mm: float = None):

        if not charuco_square_size_mm:
            charuco_square_size_mm = float(self._charuco_square_size_line_edit.text())

        active_recording_info = self._get_active_recording_info_callable()
        if active_recording_info is None:
            logger.error("Cannot calibrate from active recording - no active recording selected")
            return

        logger.info(f"Calibrating from active recording: {active_recording_info.name}")

        if not active_recording_info.synchronized_videos_status_check:
            logger.error(
                f"Cannot calibrate from {active_recording_info.name} -"
                f" `active_recording_info.synchronized_videos_status_check` is "
                f"{active_recording_info.synchronized_videos_status_check}"
            )

            return

        self._anipose_calibration_frame_worker = AniposeCalibrationThreadWorker(
            calibration_videos_folder_path=active_recording_info.synchronized_videos_folder_path,
            charuco_square_size=float(charuco_square_size_mm),
            kill_thread_event=self._kill_thread_event,
        )

        self._anipose_calibration_frame_worker.start()

        self._anipose_calibration_frame_worker.finished.connect(self.update_calibration_toml_path)


if __name__ == "__main__":
    from PyQt6.QtWidgets import QApplication
    import sys

    app = QApplication(sys.argv)
    widget = CalibrationControlPanel()
    widget.show()
    sys.exit(app.exec())



import logging

from PyQt6.QtWidgets import QVBoxLayout, QWidget
from qtconsole.manager import QtKernelManager
from qtconsole.rich_jupyter_widget import RichJupyterWidget

logger = logging.getLogger(__name__)


class JupyterConsoleWidget(QWidget):
    def __init__(self, dark_mode: bool = True, parent=None):
        super().__init__(parent=parent)

        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self._kernel_manager = QtKernelManager(kernel_name="python3")
        self._kernel_manager.start_kernel()

        self._kernel_client = self._kernel_manager.client()
        self._kernel_client.start_channels()

        self._jupyter_widget = RichJupyterWidget()
        self._jupyter_widget.kernel_manager = self._kernel_manager
        self._jupyter_widget.kernel_client = self._kernel_client

        self._layout.addWidget(self._jupyter_widget)

        self._import_stuff()

        self._connect_to_logging()

        # get_qt_app().aboutToQuit.connect(self.shutdown_kernel)

        if dark_mode:
            self._jupyter_widget.set_default_style("linux")

    @property
    def jupyter_widget(self):
        return self._jupyter_widget

    def _start_kernel_and_whatnot(self):
        """Start a kernel, connect to it, and create a RichJupyterWidget to use it"""
        kernel_manager = QtKernelManager(kernel_name="python3")
        kernel_manager.start_kernel()

        kernel_client = kernel_manager.client()
        kernel_client.start_channels()

        jupyter_widget = RichJupyterWidget()
        jupyter_widget.kernel_manager = kernel_manager
        jupyter_widget.kernel_client = kernel_client
        return jupyter_widget

    def _import_stuff(self):
        self.execute("import matplotlib.pyplot as plt", hidden=True)
        self.execute("import numpy as np", hidden=True)
        self.execute("%whos")

    def execute(self, code: str, hidden: bool = False):
        self._jupyter_widget.execute(code, hidden=hidden)

    def print_to_console(self, message: str):
        self.execute(f"print('{message}')")

    def write(self, text: str):
        self._jupyter_widget._append_plain_text(text)

    def shutdown_kernel(self):
        self._jupyter_widget.kernel_client.stop_channels()
        self._jupyter_widget.kernel_manager.shutdown_kernel()

    def _connect_to_logging(self):
        return
        # logger.info("Connecting Jupyter Console Widget to Logging")
        # logging.getLogger("").handlers.append(logging.StreamHandler(self))


if __name__ == "__main__":
    from freemocap.system.logging.configure_logging import configure_logging

    configure_logging()
    logger.info("Hello world!")

    from PyQt6.QtWidgets import QApplication

    app = QApplication([])

    jupyter_widget = JupyterConsoleWidget()
    jupyter_widget.show()

    h = logging.StreamHandler(jupyter_widget)
    logging.getLogger("").handlers.append(h)
    logger.info("Hello world worldddd!")
    logger.info("Hello world worlddd WWWWWEEEEEEE!")

    logger.info("gfsdgdfg!")
    logger.info("gfsdgdfsdfsdfsg!")
    app.exec()



import json
import logging
from pathlib import Path
from typing import Union

logger = logging.getLogger(__name__)


def save_dictionary_to_json(save_path: Union[Path, str], dictionary: dict, file_name: str):
    if file_name.split(".")[-1] != "json":
        file_name = f"{file_name}.json"

    json_file_path = Path(save_path) / file_name

    with open(json_file_path, "w") as file:
        json.dump(dictionary, file, indent=4)

    logger.info(f"Saved dictionary {file_name} as json  at: {json_file_path}")



# from -https://gist.github.com/peace098beat/db8ef7161508e6500ebe?permalink_comment_id=3542714#gistcomment-3542714

from PyQt6.QtWidgets import QMainWindow, QApplication


import sys


class MainWidget(QMainWindow):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("Drag and Drop")
        self.resize(720, 480)
        self.setAcceptDrops(True)

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        files = [u.toLocalFile() for u in event.mimeData().urls()]
        for f in files:
            print(f)


if __name__ == "__main__":
    app = QApplication(sys.argv)
    ui = MainWidget()
    ui.show()
    sys.exit(app.exec())



from pathlib import Path
from typing import Union

import numpy as np

from freemocap.tests.utilities.get_number_of_frames_of_videos_in_a_folder import (
    get_number_of_frames_of_videos_in_a_folder,
)


def test_mediapipe_3d_data_shape(
    synchronized_videos_folder: Union[str, Path],
    mediapipe_3d_data_npy_path: Union[str, Path],
    medipipe_reprojection_error_data_npy_path: Union[str, Path],
):
    """
    test that the `mediapipe 3d detection` process worked correctly by checking:

    1. There is an `.npy` file containing the mediapipe 3d data in the `output_data_folder`
    2. The dimensions of that `npy` ( number of frames, [ need to do - number of tracked points], [X,Y,Z] matches the videos in the `synchronized videos` folder

    TODO - check number of tracked points vs 'expected' number of tracked points
    """

    assert Path(
        mediapipe_3d_data_npy_path
    ).is_file(), f"3d skeleton data file does not exist at {mediapipe_3d_data_npy_path}"

    skel3d_frame_marker_xyz = np.load(mediapipe_3d_data_npy_path)

    assert (
        len(skel3d_frame_marker_xyz.shape) == 3
    ), f"3d skeleton data file should have 3 dimensions -  {mediapipe_3d_data_npy_path}"

    assert Path(
        medipipe_reprojection_error_data_npy_path
    ).is_file(), f"3d skeleton reprojection error data file does not exist at {medipipe_reprojection_error_data_npy_path}"

    skeleton_reprojection_error_fr_mar = np.load(
        medipipe_reprojection_error_data_npy_path
    )

    assert (
        len(skeleton_reprojection_error_fr_mar.shape) == 2
    ), f"3d skeleton reprojection error data file should have 2 dimensions {medipipe_reprojection_error_data_npy_path}"

    frame_count = get_number_of_frames_of_videos_in_a_folder(synchronized_videos_folder)
    assert (
        len(set(frame_count)) == 1
    ), f"Videos in {synchronized_videos_folder} have different frame counts: {frame_count}"

    number_of_frames = frame_count[0]

    assert skel3d_frame_marker_xyz.shape[0] == number_of_frames
    assert skeleton_reprojection_error_fr_mar.shape[0] == number_of_frames

    # TODO - check number of tracked points vs 'expected' number of tracked points

    assert (
        skel3d_frame_marker_xyz.shape[2] == 3
    ), f"3d skeleton data file does not have 3 dimensions for X,Y,Z at {mediapipe_3d_data_npy_path}"

    return True



import logging
import sys

from BlendArMocap.src.cgt_freemocap.fm_subprocess_cmd_receiver import import_freemocap_session

logger = logging.getLogger(__name__)

argv = sys.argv
argv = argv[argv.index("--") + 1 :]
recording_path = argv[0]
blender_save_path = argv[1]
bind_to_rig_bool = bool(argv[2])
load_synchronized_videos_bool = bool(argv[3])
timeout = bool(argv[4])
load_raw_bool = bool(argv[5])
load_quick_bool = bool(argv[6])

logger.info(
    f"Invoking CGTinker's BlendArMocap based export method with arguments -"
    f"\nrecording_path: {recording_path}, "
    f"\nblender_save_path: {blender_save_path}, "
    f"\nbind_to_rig_bool: {bind_to_rig_bool}, "
    f"\nload_synchronized_videos_bool: {load_synchronized_videos_bool}, "
    f"\ntimeout: {timeout}, "
    f"\nload_raw_bool: {load_raw_bool}, "
    f"\nload_quick_bool: {load_quick_bool}"
)

import_freemocap_session(
    session_directory=recording_path,
    bind_to_rig=bind_to_rig_bool,
    load_synch_videos=load_synchronized_videos_bool,
    timeout=timeout,
    load_raw=load_raw_bool,
    load_quick=load_quick_bool,
)



# %%
import logging
import sys
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
from rich.progress import track
from scipy import signal

from freemocap.core_processes.detecting_things_in_2d_images.mediapipe_stuff.mediapipe_skeleton_names_and_connections import \
    NUMBER_OF_MEDIAPIPE_BODY_MARKERS
from freemocap.system.paths_and_files_names import (
    CENTER_OF_MASS_FOLDER_NAME,
    SEGMENT_CENTER_OF_MASS_NPY_FILE_NAME,
    TOTAL_BODY_CENTER_OF_MASS_NPY_FILE_NAME,
)

logger = logging.getLogger(__name__)


def find_good_clean_frame_reprojection_error_method(skeleton_3d_fr_mar_xyz, skeleton_reprojection_error_fr_mar):
    """
    Find the frame with the fewest nans, scaled by reprojection error (i.e. hopefully a frame where all tracked points are visible and well tracked)
    """
    print(
        "estimating good clean frame as the one where there few `nans`, reprojection error is low, and velocity is low..."
    )

    nans_per_frame = np.sum(np.isnan(skeleton_3d_fr_mar_xyz[:, :, 0]), axis=1)
    reprojection_error_per_frame = np.nanmedian(skeleton_reprojection_error_fr_mar, axis=1)
    marker_velocity_per_frame = np.nanmedian(np.abs(np.diff(skeleton_3d_fr_mar_xyz, axis=1)), axis=1)
    marker_velocity_per_frame = np.nanmean(np.abs(marker_velocity_per_frame), axis=1)

    print(f"nans_per_frame.shape: {nans_per_frame.shape}")
    print(f"reprojection_error_per_frame.shape: {reprojection_error_per_frame.shape}")
    print(f"marker_velocity_per_frame.shape: {marker_velocity_per_frame.shape}")

    # normalize errors by relevant thing
    nans_per_frame_normalized = nans_per_frame / skeleton_3d_fr_mar_xyz.shape[1]
    reprojection_error_per_frame_normalized = reprojection_error_per_frame / np.nanmedian(reprojection_error_per_frame)
    marker_velocity_per_frame_normalized = marker_velocity_per_frame / np.nanmedian(marker_velocity_per_frame)

    print(
        f"nans_per_frame_normalized (mean, [standard_deviation] : {np.nanmean(nans_per_frame_normalized):.3f} [{np.nanstd(nans_per_frame_normalized):.3f}]"
    )
    print(
        f"reprojection_error_per_frame_normalized (mean, [standard_deviation] : {np.nanmean(reprojection_error_per_frame_normalized):.3f} [{np.nanstd(reprojection_error_per_frame_normalized):.3f}]"
    )
    print(
        f"marker_velocity_per_frame_normalized (mean, [standard_deviation] : {np.nanmean(marker_velocity_per_frame_normalized):.3f} [{np.nanstd(marker_velocity_per_frame_normalized):.3f}]"
    )

    frame_cleanliness_score_where_low_scores_are_good = (
            nans_per_frame_normalized + reprojection_error_per_frame_normalized + marker_velocity_per_frame_normalized
    )

    good_clean_frame_number = np.nanargmin(frame_cleanliness_score_where_low_scores_are_good)
    print(f"----estimated good_clean_frame_number: {good_clean_frame_number}----")
    return good_clean_frame_number


# %%
def fill_gaps_in_freemocap_data(inputData_frame_marker_xyz: np.ndarray) -> np.ndarray:
    """
    Takes in a 3d skeleton numpy array from freemocap and interpolates missing NaN values
    TODO - refactor so it doesn't use a for loop (and maybe uses more sophisticated gap-fill methods?)
    """

    interpolatedData_frame_marker_xyz = inputData_frame_marker_xyz.copy()

    for marker_number in range(interpolatedData_frame_marker_xyz.shape[1]):
        trajectory_frame_xyz = inputData_frame_marker_xyz[:, marker_number, :]

        if marker_number < NUMBER_OF_MEDIAPIPE_BODY_MARKERS:
            df = pd.DataFrame(trajectory_frame_xyz).interpolate(method="linear", axis=0)
            df.fillna(method="bfill", axis=0, inplace=True)
            df.fillna(method="ffill", axis=0, inplace=True)

            # if np.any(np.isnan(df.to_numpy())):
            #     raise ValueError("Nans still present after interpolation")
            # if np.any(np.isinf(df.to_numpy())):
            #     raise ValueError("Infs still present after interpolation")

            interpolatedData_frame_marker_xyz[:, marker_number, :] = df.to_numpy()

    return interpolatedData_frame_marker_xyz


# %%
def butterworth_lowpass_zerolag_filter(data, cutoff, sampling_rate, order):
    """Run a low pass butterworth filter on a single column of data"""
    nyquist_freq = 0.5 * sampling_rate
    normal_cutoff = cutoff / nyquist_freq
    # Get the filter coefficients
    b, a = signal.butter(order, normal_cutoff, btype="low", analog=False)
    y = signal.filtfilt(b, a, data)
    return y


def butterworth_filter_skeleton(skeleton_frame_marker_xyz, cutoff, sampling_rate, order):
    """Take in a 3d skeleton numpy array and calculate_center_of_mass a low pass butterworth filter on each marker in the data"""

    butterworth_filtered_data = skeleton_frame_marker_xyz.copy()

    for marker_number in range(NUMBER_OF_MEDIAPIPE_BODY_MARKERS):
        trajectory_xyz = skeleton_frame_marker_xyz[:, marker_number, :]
        if np.any(np.isnan(trajectory_xyz)):
            logger.warning(f"Trajectory number {marker_number} has nans, skipping butterworth filter")
            continue
        for dimension in range(trajectory_xyz.shape[1]):
            butterworth_filtered_data[:, marker_number, dimension] = butterworth_lowpass_zerolag_filter(
                trajectory_xyz[:,dimension],
                cutoff,
                sampling_rate,
                order,
            )

    assert skeleton_frame_marker_xyz.shape == butterworth_filtered_data.shape

    return butterworth_filtered_data


# %%
def find_velocity_values_within_limit(skeleton_velocity_data, velocity_limit):
    """
    This function takes in a skeleton velocity data array and a limit and returns the indices of the values that are within the limit
    """
    indices = []
    for i in range(len(skeleton_velocity_data)):
        if abs(skeleton_velocity_data[i]) <= velocity_limit:
            indices.append(
                i + 1
            )  # add 1 to account for the difference in indices between the position and velocity data
    return indices


# %%


def find_matching_indices_in_lists(list_1, list_2, list_3, list_4):
    """
    This function takes in four lists and returns the indices of the values that are in all four lists
    """
    matching_values = [x for x in list_1 if x in list_2 and x in list_3 and x in list_4]

    return matching_values


# %%


def find_best_velocity_guess(skeleton_velocity_data, skeleton_indices, velocity_guess, iteration_range):
    """
    This function iterates over velocity data and tries to pare down to a single frame that has the closest velocity to 0 for all foot markers
    """

    print(f"Velocity guess: {velocity_guess}, iteration range: {iteration_range}")

    right_heel_index = skeleton_indices.index("right_heel")
    right_toe_index = skeleton_indices.index("right_foot_index")
    left_heel_index = skeleton_indices.index("left_heel")
    left_toe_index = skeleton_indices.index("left_foot_index")

    skeleton_data_velocity_x_right_heel = skeleton_velocity_data[:, right_heel_index, 0]
    skeleton_data_velocity_x_right_toe = skeleton_velocity_data[:, right_toe_index, 0]
    skeleton_data_velocity_x_left_heel = skeleton_velocity_data[:, left_heel_index, 0]
    skeleton_data_velocity_x_left_toe = skeleton_velocity_data[:, left_toe_index, 0]

    print(
        f"sum nan right heel: {np.sum(np.isnan(skeleton_data_velocity_x_right_heel))}, "
        f"\nsum nan right toe: {np.sum(np.isnan(skeleton_data_velocity_x_right_toe))}, "
        f"\nsum nan left heel: {np.sum(np.isnan(skeleton_data_velocity_x_left_heel))}, "
        f"\nsum nan left toe: {np.sum(np.isnan(skeleton_data_velocity_x_left_toe))}, "
        f"\nnum frames: {len(skeleton_data_velocity_x_right_heel)}"
    )

    # get a list of the frames where the velocity for that marker is within the velocity limit
    right_heel_x_velocity_limits = find_velocity_values_within_limit(
        skeleton_data_velocity_x_right_heel, velocity_guess
    )
    right_toe_x_velocity_limits = find_velocity_values_within_limit(skeleton_data_velocity_x_right_toe, velocity_guess)
    left_heel_x_velocity_limits = find_velocity_values_within_limit(skeleton_data_velocity_x_left_heel, velocity_guess)
    left_toe_x_velocity_limits = find_velocity_values_within_limit(skeleton_data_velocity_x_left_toe, velocity_guess)

    # return a list of matching frame indices from the four lists generated above
    matching_values = find_matching_indices_in_lists(
        right_heel_x_velocity_limits,
        right_toe_x_velocity_limits,
        left_heel_x_velocity_limits,
        left_toe_x_velocity_limits,
    )
    matching_values = [x for x in matching_values if x > 75]

    # print(matching_values)
    if len(matching_values) > 1 and velocity_guess > 0:
        # if there are multiple matching values, decrease the guess a little bit and calculate_center_of_mass the function again
        #
        velocity_guess = velocity_guess - iteration_range
        print(
            "Current Velocity Guess:",
            velocity_guess,
            "| Number of Possible Frames:",
            len(matching_values),
            "| Possible Frames:",
            matching_values,
        )
        matching_values, velocity_guess = find_best_velocity_guess(
            skeleton_velocity_data, skeleton_indices, velocity_guess, iteration_range
        )

        f = 2
    elif len(matching_values) == 0:
        # if there are no matching values (we decreased our guess too far), reset the guess to be a bit smaller and calculate_center_of_mass the function again with smaller intervals between the guesses
        iteration_range = iteration_range / 2
        matching_values, velocity_guess = find_best_velocity_guess(
            skeleton_velocity_data,
            skeleton_indices,
            velocity_guess + iteration_range * 2,
            iteration_range,
        )

        f = 2
    elif len(matching_values) == 1:
        print("Good Frame:", matching_values, "| Final Velocity Guess:", velocity_guess)

    return matching_values, velocity_guess


# %%


def find_good_frame_recursive_guess_method(
        skeleton_data, skeleton_indices: list, initial_velocity_guess: float, debug=False
):
    """
    Finds a frame (called the good frame) where the velocity of both feet are closest to 0

    Input:
        skeleton data: a 3D numpy array of skeleton data in freemocap format
        skeleton indices: a list of joints being tracked by mediapipe/your 2d pose estimator
        initial velocity guess: just a starting guess for the optimizer. Can adjust if you're not getting the results you want
        debug: plots and displays the calculated good frame if True
    """

    skeleton_velocity_data = np.diff(skeleton_data, axis=0)
    print("finding best velocity guess...")
    matching_values, velocity_guess = find_best_velocity_guess(
        skeleton_velocity_data,
        skeleton_indices,
        initial_velocity_guess,
        iteration_range=0.1,
    )
    print(f"Return values: {matching_values}")
    good_frame = matching_values[0]

    return good_frame


# %%


def create_vector(point1, point2):
    """Put two points in, make a vector"""
    vector = point2 - point1
    return vector


# %%


def normalize_vector(vector):
    """Take in a vector, normalize it"""
    unit_vector = vector / np.linalg.norm(vector)

    vector_length = np.sqrt(np.sum(unit_vector ** 2))
    assert (
            np.abs(vector_length - 1.0) < sys.float_info.epsilon
    ), f"Vector is not normalized: vector_length = {vector_length}"
    return unit_vector


# %%


def calculate_shoulder_center_XYZ_coordinates(single_frame_skeleton_data, left_shoulder_index, right_shoulder_index):
    """Take in the left and right shoulder indices, and calculate the shoulder center point"""
    left_shoulder_point = single_frame_skeleton_data[left_shoulder_index, :]
    right_shoulder_point = single_frame_skeleton_data[right_shoulder_index, :]
    shoulder_center_XYZ_coordinates = (left_shoulder_point + right_shoulder_point) / 2

    return shoulder_center_XYZ_coordinates


# %%


def calculate_mid_hip_XYZ_coordinates(single_frame_skeleton_data, left_hip_index, right_hip_index):
    """Take in the left and right hip indices, and calculate the mid hip point"""
    left_hip_point = single_frame_skeleton_data[left_hip_index, :]
    right_hip_point = single_frame_skeleton_data[right_hip_index, :]
    mid_hip_XYZ_coordinates = (left_hip_point + right_hip_point) / 2

    return mid_hip_XYZ_coordinates


# %%


def calculate_mid_foot_XYZ_coordinate(
        single_frame_skeleton_data,
        left_heel_index,
        right_heel_index,
):
    """Take in the primary and secondary foot indices, and calculate the mid foot point"""
    right_foot_point = single_frame_skeleton_data[right_heel_index, :]
    left_foot_point = single_frame_skeleton_data[left_heel_index, :]
    mid_foot_XYZ_coordinates = (right_foot_point + left_foot_point) / 2

    return mid_foot_XYZ_coordinates


# def calculate_translation_distance(skeleton_point_coordinate):
#     """Take a skeleton point coordinate and calculate its distance to the origin"""

#     translation_distance = skeleton_point_coordinate - [0,0,0]
#     return translation_distance


# %%


def translate_skeleton_frame(rotated_skeleton_data_frame, translation_distance):
    """Take in a frame of rotated skeleton data, and apply the translation distance to each point in the skeleton"""

    translated_skeleton_frame = rotated_skeleton_data_frame - translation_distance
    return translated_skeleton_frame


# %%


def translate_skeleton_to_origin(point_to_translate, original_skeleton_data):
    num_frames = original_skeleton_data.shape[0]

    translated_skeleton_data = np.zeros(original_skeleton_data.shape)

    for frame in track(range(num_frames), description="Translating Skeleton"):
        translated_skeleton_data[frame, :, :] = translate_skeleton_frame(
            original_skeleton_data[frame, :, :], point_to_translate
        )

    return translated_skeleton_data


# %%


def calculate_skewed_symmetric_cross_product(cross_product_vector):
    # needed in the calculate_rotation_matrix function
    skew_symmetric_cross_product = np.array(
        [
            [0, -cross_product_vector[2], cross_product_vector[1]],
            [cross_product_vector[2], 0, -cross_product_vector[0]],
            [-cross_product_vector[1], cross_product_vector[0], 0],
        ]
    )
    return skew_symmetric_cross_product


# %%


def calculate_rotation_matrix(vector1, vector2):
    """Put in two vectors to calculate the rotation matrix between those two vectors"""
    # based on the code found here: https://math.stackexchange.com/questions/180418/calculate-rotation-matrix-to-align-vector-a-to-vector-b-in-3d"""

    identity_matrix = np.identity(3)
    vector_cross_product = np.cross(vector1, vector2)
    vector_dot_product = np.dot(vector1, vector2)
    skew_symmetric_cross_product = calculate_skewed_symmetric_cross_product(vector_cross_product)
    rotation_matrix = (
            identity_matrix
            + skew_symmetric_cross_product
            + (np.dot(skew_symmetric_cross_product, skew_symmetric_cross_product))
            * (1 - vector_dot_product)
            / (np.linalg.norm(vector_cross_product) ** 2)
    )

    return rotation_matrix


# %%


def rotate_point(point, rotation_matrix):
    rotated_point = np.dot(rotation_matrix, point)
    return rotated_point


# %%


def rotate_skeleton_frame(this_frame_aligned_skeleton_data, rotation_matrix):
    """Take in a frame of skeleton data, and apply the rotation matrix to each point in the skeleton"""

    this_frame_rotated_skeleton = np.zeros(
        this_frame_aligned_skeleton_data.shape
    )  # initialize the array to hold the rotated skeleton data for this frame
    num_tracked_points = this_frame_aligned_skeleton_data.shape[0]

    for i in range(num_tracked_points):
        this_frame_rotated_skeleton[i, :] = rotate_point(this_frame_aligned_skeleton_data[i, :], rotation_matrix)

    return this_frame_rotated_skeleton


# %%


def calculate_skeleton_rotation_matrix(
        skeleton_right_vector: np.ndarray,
        skeleton_up_vector: np.ndarray,
        skeleton_forward_vector: np.ndarray,
        world_up_vector: str,
):
    """Calculate the rotation matrix to align the skeleton to the world coordinate system"""

    skel_positive_x = normalize_vector(skeleton_right_vector)
    skel_positive_y = normalize_vector(skeleton_forward_vector)
    skel_positive_z = normalize_vector(skeleton_up_vector)

    z_hat = np.cross(skel_positive_x, skel_positive_y)
    y_hat = np.cross(z_hat, skel_positive_x)
    x_hat = np.cross(y_hat, z_hat)
    z_hat = np.cross(x_hat, y_hat)
    y_hat = np.cross(z_hat, x_hat)

    if world_up_vector == "z":  # biomechanics convention (assumes 'groundplane' is primary XY plane)
        rotation_matrix = np.array([x_hat, y_hat, z_hat])
    elif world_up_vector == "y":  # computer graphics convention (assumes 'viewingplane' is primary XY plane)
        rotation_matrix = np.array([x_hat, z_hat, y_hat])
    elif world_up_vector == "x":  # I don't know of anyone who uses X as vertical, but you do you boo
        rotation_matrix = np.array([z_hat, y_hat, x_hat])
    else:
        raise ValueError("world_up_vector must be one of 'x', 'y', or 'z'")

    assert (
            np.dot(rotation_matrix[0, :], rotation_matrix[1, :]) < sys.float_info.epsilon
    ), "x_hat and y_hat are not orthogonal"
    assert (
            np.dot(rotation_matrix[1, :], rotation_matrix[2, :]) < sys.float_info.epsilon
    ), "y_hat and z_hat are not orthogonal"
    assert (
            np.dot(rotation_matrix[2, :], rotation_matrix[0, :]) < sys.float_info.epsilon
    ), "z_hat and x_hat are not orthogonal"

    return rotation_matrix


def rotate_align_skeleton(
        positive_x_reference_vector_on_the_skeleton: np.ndarray,
        positive_y_reference_vector_on_the_skeleton: np.ndarray,
        positive_z_reference_vector_on_the_skeleton: np.ndarray,
        world_up_vector: str,
        original_skeleton_np_array: np.ndarray,
) -> np.ndarray:
    """
    Find the rotation matrix needed to rotate the 'reference_vector_on_the_skeleton' to match the 'world_vector_to_align_skeleton_vector_to', and
    rotate the entire skeleton with that matrix.

        Input:
            reference_vector_on_the_skeleton: The vector on the skeleton that you want to rotate/base the rotation matrix on
            world_vector_to_align_skeleton_vector_to: The vector that you want to align the skeleton too (i.e. the x-axis/y-axis etc.)
            Original skeleton data: The freemocap data you want to rotate
        Output:
            rotated_skeleton_data: A numpy data array of your rotated skeleton

    """

    skeleton_original_rotation_matrix = calculate_skeleton_rotation_matrix(
        skeleton_right_vector=positive_x_reference_vector_on_the_skeleton,
        skeleton_up_vector=positive_z_reference_vector_on_the_skeleton,
        skeleton_forward_vector=positive_y_reference_vector_on_the_skeleton,
        world_up_vector=world_up_vector,
    )
    logger.info(f"Skeleton original rotation matrix: \n {str(skeleton_original_rotation_matrix)}")
    skeleton_inverse_rotation_matrix = np.linalg.inv(skeleton_original_rotation_matrix)

    logger.info(f"Skeleton inverse rotation matrix: \n {skeleton_inverse_rotation_matrix}")

    skeleton_z_vector_rotated = positive_z_reference_vector_on_the_skeleton @ skeleton_inverse_rotation_matrix
    original_z_vector_length = np.sqrt(np.sum(positive_z_reference_vector_on_the_skeleton ** 2))
    rotated_up_vector_length = np.sqrt(np.sum(skeleton_z_vector_rotated ** 2))

    assert (
            original_z_vector_length - rotated_up_vector_length < 1
    ), f"Rotating the skeleton up vector changed its size -\ something went wrong (original_z_vector_length {original_z_vector_length} - rotated_up_vector_length {rotated_up_vector_length} = {original_z_vector_length - rotated_up_vector_length})"

    skeleton_left_vector_rotated = positive_x_reference_vector_on_the_skeleton @ skeleton_inverse_rotation_matrix

    rotated_left_vector_length = np.sqrt(np.sum(skeleton_left_vector_rotated ** 2))
    original_x_vector_length = np.sqrt(np.sum(positive_x_reference_vector_on_the_skeleton ** 2))

    assert (
            rotated_left_vector_length - original_x_vector_length < 1
    ), f"Rotating the skeleton left vector changed its size - something went wrong (rotated_left_vector_length {rotated_left_vector_length} - original_x_vector_length {original_x_vector_length} = {rotated_left_vector_length - original_x_vector_length})"

    print(
        f"Skeleton up vector (rounded) - original: {np.round(positive_z_reference_vector_on_the_skeleton)}, rotated: {np.round(skeleton_z_vector_rotated)}"
    )
    print(
        f"Skeleton left vector (rounded) - original: {np.round(positive_x_reference_vector_on_the_skeleton)}, rotated: {np.round(skeleton_left_vector_rotated)}"
    )

    rotated_skeleton_data_array = np.empty(original_skeleton_np_array.shape)
    for frame_number in track(range(original_skeleton_np_array.shape[0]), description="Rotating Skeleton..."):
        for point_number in range(original_skeleton_np_array.shape[1]):
            rotated_point_xyz = (
                    original_skeleton_np_array[frame_number, point_number, :] @ skeleton_inverse_rotation_matrix
            )

            rotated_skeleton_data_array[frame_number, point_number, :] = rotated_point_xyz

    return rotated_skeleton_data_array


# %%
def align_skeleton_with_origin_mean_blob_method(
        skeleton_3d_frame_landmark_xyz: np.ndarray,
) -> np.ndarray:
    """get stats of skeleton postions and align with origin and point upwards based on that"""
    skeleton_3d_frame_landmark_xyz_og = skeleton_3d_frame_landmark_xyz.copy()

    # get the mean of the skeleton
    skel_median_x = np.nanmedian(skeleton_3d_frame_landmark_xyz[:, :, 0])
    skel_median_y = np.nanmedian(skeleton_3d_frame_landmark_xyz[:, :, 1])
    skel_median_z = np.nanmedian(skeleton_3d_frame_landmark_xyz[:, :, 2])

    # translate the skeleton to the origin
    skeleton_3d_frame_landmark_xyz[:, :, 0] = skeleton_3d_frame_landmark_xyz[:, :, 0] - skel_median_x
    skeleton_3d_frame_landmark_xyz[:, :, 1] = skeleton_3d_frame_landmark_xyz[:, :, 1] - skel_median_y
    skeleton_3d_frame_landmark_xyz[:, :, 2] = skeleton_3d_frame_landmark_xyz[:, :, 2] - skel_median_z

    skeleton_3d_frame_landmark_xyz_translated = skeleton_3d_frame_landmark_xyz.copy()

    # pull out head markers to define 'up'
    # (mediapipe - frames 0-10 are the head)
    head_median_x = np.nanmedian(skeleton_3d_frame_landmark_xyz[:10, :, 0])
    head_median_y = np.nanmedian(skeleton_3d_frame_landmark_xyz[:10, :, 1])
    head_median_z = np.nanmedian(skeleton_3d_frame_landmark_xyz[:10, :, 2])

    skeleton_up_vector = np.array([head_median_x, head_median_y, head_median_z])
    world_up_vector = np.array([0, 0, 1])

    # rotate the skeleton to point upwards
    skeleton_3d_frame_landmark_xyz = rotate_align_skeleton(
        reference_vector_on_the_skeleton=skeleton_up_vector,
        world_vector_to_align_skeleton_vector_to=world_up_vector,
        original_skeleton_np_array=skeleton_3d_frame_landmark_xyz,
    )

    return skeleton_3d_frame_landmark_xyz


def align_skeleton_with_origin_foot_spine_method(
        skeleton_data: np.ndarray, skeleton_indices: list, good_frame: int
) -> np.ndarray:
    """
    Takes in freemocap skeleton data and translates the skeleton to the origin, and then rotates the data
    so that the skeleton is facing the +y direction and standing in the +z direction

    Input:
        skeleton data: a 3D numpy array of skeleton data in freemocap format
        skeleton indices: a list of joints being tracked by mediapipe/your 2d pose estimator
        good frame: the frame that you want to base the rotation on (can be entered manually,
                    or use the 'good_frame_finder.py' to calculate it)
        debug: If 'True', display a plot of the raw data and the 3 main alignment stages

    Output:
        spine aligned skeleton data: a 3d numpy array of the origin aligned data in freemocap format
    """

    ## Translate the data such that the midpoint between the two feet is at the origin

    good_skeleton_frame = skeleton_data[good_frame, :, :]

    # the mid point between the two heels
    midpoint_between_heels_xyz = np.mean(
        good_skeleton_frame[
        [skeleton_indices.index("left_heel"), skeleton_indices.index("right_heel")],
        :,
        ],
        axis=0,
    )

    # subtract the mean heel  position from the skeleton data (so it is at the origin, )
    skeleton_data_zeroed = skeleton_data.copy()
    skeleton_data_zeroed[:, :, 0] -= midpoint_between_heels_xyz[0]
    skeleton_data_zeroed[:, :, 1] -= midpoint_between_heels_xyz[1]
    skeleton_data_zeroed[:, :, 2] -= midpoint_between_heels_xyz[2]

    # rotate so left heel is on the +Y axis
    right_heel_vector = skeleton_data_zeroed[good_frame, skeleton_indices.index("right_heel"), :]

    # Rotating the skeleton so that the spine is aligned with +z
    midpoint_between_hips_xyz = np.mean(
        skeleton_data_zeroed[
        good_frame,
        [skeleton_indices.index("left_hip"), skeleton_indices.index("right_hip")],
        :,
        ],
        axis=0,
    )

    midpoint_between_toes_xyz = np.mean(
        skeleton_data_zeroed[
        good_frame,
        [
            skeleton_indices.index("left_foot_index"),
            skeleton_indices.index("right_foot_index"),
        ],
        :,
        ],
        axis=0,
    )

    origin_aligned_skeleton_data = rotate_align_skeleton(
        positive_x_reference_vector_on_the_skeleton=right_heel_vector,
        positive_y_reference_vector_on_the_skeleton=midpoint_between_toes_xyz,
        positive_z_reference_vector_on_the_skeleton=midpoint_between_hips_xyz,
        world_up_vector="z",
        original_skeleton_np_array=skeleton_data_zeroed,
    )

    return origin_aligned_skeleton_data


# %%
mediapipe_landmark_names = [
    "nose",
    "left_eye_inner",
    "left_eye",
    "left_eye_outer",
    "right_eye_inner",
    "right_eye",
    "right_eye_outer",
    "left_ear",
    "right_ear",
    "mouth_left",
    "mouth_right",
    "left_shoulder",
    "right_shoulder",
    "left_elbow",
    "right_elbow",
    "left_wrist",
    "right_wrist",
    "left_pinky",
    "right_pinky",
    "left_index",
    "right_index",
    "left_thumb",
    "right_thumb",
    "left_hip",
    "right_hip",
    "left_knee",
    "right_knee",
    "left_ankle",
    "right_ankle",
    "left_heel",
    "right_heel",
    "left_foot_index",
    "right_foot_index",
]


def return_indices_of_joints(list_of_indices, list_of_joint_names):
    indices = []
    for name in list_of_joint_names:
        this_name_index = list_of_indices.index(name)
        indices.append(this_name_index)

    return indices


# %%


def return_XYZ_coordinates_of_markers(freemocap_data, indices_list, frame):
    XYZ_coordinates = []
    for index in indices_list:
        this_joint_coordinate = freemocap_data[frame, index, :]
        XYZ_coordinates.append(this_joint_coordinate)

    return XYZ_coordinates


# %%


def build_virtual_trunk_marker(freemocap_data, list_of_indices, trunk_joint_connection, frame):
    trunk_marker_indices = return_indices_of_joints(list_of_indices, trunk_joint_connection)

    trunk_XYZ_coordinates = return_XYZ_coordinates_of_markers(freemocap_data, trunk_marker_indices, frame)

    trunk_proximal = (trunk_XYZ_coordinates[0] + trunk_XYZ_coordinates[1]) / 2
    trunk_distal = (trunk_XYZ_coordinates[2] + trunk_XYZ_coordinates[3]) / 2

    return trunk_proximal, trunk_distal


# %%


def build_mediapipe_skeleton(mediapipe_pose_data, segment_dataframe, mediapipe_indices) -> list:
    """This function takes in the mediapipe pose data array and the segment_conn_len_perc_dataframe.
    For each frame of data, it loops through each segment we want to find and identifies the names
    of the proximal and distal joints of that segment. Then it searches the mediapipe_indices list
    to find the index that corresponds to the name of that segment. We plug the index into the
    mediapipe_pose_data array to find the proximal/distal joints' XYZ coordinates at that frame.
    The segment, its proximal joint and its distal joint gets thrown into a dictionary.
    And then that dictionary gets saved to a list for each frame. By the end of the function, you
    have a list that contains the skeleton segment XYZ coordinates for each frame."""

    num_frames = mediapipe_pose_data.shape[0]
    num_frame_range = range(num_frames)

    mediapipe_frame_segment_joint_XYZ = []  # empty list to hold all the skeleton XYZ coordinates/frame

    for frame in track(num_frame_range, description="Building a MediaPipe Skeleton"):

        trunk_joint_connection = [
            "left_shoulder",
            "right_shoulder",
            "left_hip",
            "right_hip",
        ]
        trunk_virtual_markers = build_virtual_trunk_marker(
            mediapipe_pose_data, mediapipe_indices, trunk_joint_connection, frame
        )

        mediapipe_pose_skeleton_coordinates = {}
        for (
                segment,
                segment_info,
        ) in (
                segment_dataframe.iterrows()
        ):  # iterate through the data frame by the segment name and all the info for that segment
            if segment == "trunk":

                # based on index, excract coordinate data from fmc mediapipe data
                mediapipe_pose_skeleton_coordinates[segment] = [
                    trunk_virtual_markers[0],
                    trunk_virtual_markers[1],
                ]
            elif segment == "left_hand" or segment == "right_hand":
                proximal_joint_hand = segment_info["Joint_Connection"][0]
                if segment == "left_hand":
                    distal_joint_hand = "left_index"
                else:
                    distal_joint_hand = "right_index"

                proximal_joint_hand_index = mediapipe_indices.index(proximal_joint_hand)
                distal_joint_hand_index = mediapipe_indices.index(distal_joint_hand)

                mediapipe_pose_skeleton_coordinates[segment] = [
                    mediapipe_pose_data[frame, proximal_joint_hand_index, :],
                    mediapipe_pose_data[frame, distal_joint_hand_index, :],
                ]

            elif segment == "left_foot" or segment == "right_foot":
                if segment == "left_foot":
                    proximal_joint_foot_name = "left_ankle"
                else:
                    proximal_joint_foot_name = "right_ankle"

                proximal_joint_foot_index = mediapipe_indices.index(proximal_joint_foot_name)

                distal_joint_foot = segment_info["Joint_Connection"][1]
                distal_joint_foot_index = mediapipe_indices.index(distal_joint_foot)
                mediapipe_pose_skeleton_coordinates[segment] = [
                    mediapipe_pose_data[frame, proximal_joint_foot_index, :],
                    mediapipe_pose_data[frame, distal_joint_foot_index, :],
                ]

            else:
                proximal_joint_name = segment_info["Joint_Connection"][0]
                distal_joint_name = segment_info["Joint_Connection"][1]

                # get the mediapipe index for the proximal and distal joint for this segment
                proximal_joint_index = mediapipe_indices.index(proximal_joint_name)
                distal_joint_index = mediapipe_indices.index(distal_joint_name)

                # use the mediapipe indices to get the XYZ coordinates for the prox/distal joints and throw it in a dictionary
                # mediapipe_pose_skeleton_coordinates[segment] = {'proximal':mediapipe_pose_data[frame,proximal_joint_index,:],'distal':mediapipe_pose_data[frame,distal_joint_index,:]}
                mediapipe_pose_skeleton_coordinates[segment] = [
                    mediapipe_pose_data[frame, proximal_joint_index, :],
                    mediapipe_pose_data[frame, distal_joint_index, :],
                ]

        mediapipe_frame_segment_joint_XYZ.append(mediapipe_pose_skeleton_coordinates)
        f = 2

    return mediapipe_frame_segment_joint_XYZ


# %%
# values for segment weight and segment mass percentages taken from Winter anthropometry tables
# https://imgur.com/a/aD74j
# Winter, D.A. (2005) Biomechanics and Motor Control of Human Movement. 3rd Edition, John Wiley & Sons, Inc., Hoboken.


segments = [
    "head",
    "trunk",
    "right_upper_arm",
    "left_upper_arm",
    "right_forearm",
    "left_forearm",
    "right_hand",
    "left_hand",
    "right_thigh",
    "left_thigh",
    "right_shin",
    "left_shin",
    "right_foot",
    "left_foot",
]

joint_connections = [
    ["left_ear", "right_ear"],
    ["mid_chest_marker", "mid_hip_marker"],
    ["right_shoulder", "right_elbow"],
    ["left_shoulder", "left_elbow"],
    ["right_elbow", "right_wrist"],
    ["left_elbow", "left_wrist"],
    ["right_wrist", "right_hand_marker"],
    ["left_wrist", "left_hand_marker"],
    ["right_hip", "right_knee"],
    ["left_hip", "left_knee"],
    ["right_knee", "right_ankle"],
    ["left_knee", "left_ankle"],
    ["right_back_of_foot_marker", "right_foot_index"],
    ["left_back_of_foot_marker", "left_foot_index"],
]

segment_COM_lengths = [
    0.5,
    0.5,
    0.436,
    0.436,
    0.430,
    0.430,
    0.506,
    0.506,
    0.433,
    0.433,
    0.433,
    0.433,
    0.5,
    0.5,
]

segment_COM_percentages = [
    0.081,
    0.497,
    0.028,
    0.028,
    0.016,
    0.016,
    0.006,
    0.006,
    0.1,
    0.1,
    0.0465,
    0.0465,
    0.0145,
    0.0145,
]


# %%
def calculate_segment_COM(
        segment_conn_len_perc_dataframe,
        skelcoordinates_frame_segment_joint_XYZ,
        num_frame_range,
):
    segment_COM_frame_dict = []
    for frame in track(num_frame_range, description="Calculating Segment Center of Mass"):
        segment_COM_dict = {}
        for segment, segment_info in segment_conn_len_perc_dataframe.iterrows():
            this_segment_XYZ = skelcoordinates_frame_segment_joint_XYZ[frame][segment]

            # for mediapipe
            this_segment_proximal = this_segment_XYZ[0]
            this_segment_distal = this_segment_XYZ[1]
            this_segment_COM_length = segment_info["Segment_COM_Length"]

            this_segment_COM = this_segment_proximal + this_segment_COM_length * (
                    this_segment_distal - this_segment_proximal
            )
            segment_COM_dict[segment] = this_segment_COM
        segment_COM_frame_dict.append(segment_COM_dict)
    return segment_COM_frame_dict


# %%


def reformat_segment_COM(segment_COM_frame_dict, num_frame_range, num_segments):
    segment_COM_frame_imgPoint_XYZ = np.empty([int(len(num_frame_range)), int(num_segments), 3])
    for frame in num_frame_range:
        this_frame_skeleton = segment_COM_frame_dict[frame]
        for joint_count, segment in enumerate(this_frame_skeleton.keys()):
            segment_COM_frame_imgPoint_XYZ[frame, joint_count, :] = this_frame_skeleton[segment]
    return segment_COM_frame_imgPoint_XYZ


# %%


def calculate_total_body_COM(segment_conn_len_perc_dataframe, segment_COM_frame_dict, num_frame_range):
    logger.info("Calculating Total Body Center of Mass")
    totalBodyCOM_frame_XYZ = np.empty([int(len(num_frame_range)), 3])

    for frame in track(num_frame_range, description="Calculating Total Body Center of Mass..."):

        this_frame_total_body_percentages = []
        this_frame_skeleton = segment_COM_frame_dict[frame]

        for segment, segment_info in segment_conn_len_perc_dataframe.iterrows():
            this_segment_COM = this_frame_skeleton[segment]
            this_segment_COM_percentage = segment_info["Segment_COM_Percentage"]

            this_segment_total_body_percentage = this_segment_COM * this_segment_COM_percentage
            this_frame_total_body_percentages.append(this_segment_total_body_percentage)

        this_frame_total_body_COM = np.nansum(this_frame_total_body_percentages, axis=0)

        totalBodyCOM_frame_XYZ[frame, :] = this_frame_total_body_COM

    f = 2
    return totalBodyCOM_frame_XYZ


# %%


def build_anthropometric_dataframe(
        segments: list,
        joint_connections: list,
        segment_COM_lengths: list,
        segment_COM_percentages: list,
) -> pd.DataFrame:
    # load anthropometric data into a pandas dataframe
    df = pd.DataFrame(
        list(
            zip(
                segments,
                joint_connections,
                segment_COM_lengths,
                segment_COM_percentages,
            )
        ),
        columns=[
            "Segment_Name",
            "Joint_Connection",
            "Segment_COM_Length",
            "Segment_COM_Percentage",
        ],
    )
    segment_conn_len_perc_dataframe = df.set_index("Segment_Name")
    return segment_conn_len_perc_dataframe


def calculate_center_of_mass(
        freemocap_marker_data_array: np.ndarray,
        pose_estimation_skeleton: list,
        anthropometric_info_dataframe: pd.DataFrame,
):
    num_frames = freemocap_marker_data_array.shape[0]
    num_frame_range = range(num_frames)
    num_segments = len(anthropometric_info_dataframe)

    segment_COM_frame_dict = calculate_segment_COM(
        anthropometric_info_dataframe, pose_estimation_skeleton, num_frame_range
    )
    segment_COM_frame_imgPoint_XYZ = reformat_segment_COM(segment_COM_frame_dict, num_frame_range, num_segments)
    totalBodyCOM_frame_XYZ = calculate_total_body_COM(
        anthropometric_info_dataframe, segment_COM_frame_dict, num_frame_range
    )

    return (
        segment_COM_frame_dict,
        segment_COM_frame_imgPoint_XYZ,
        totalBodyCOM_frame_XYZ,
    )


# %%
def are_there_feet_in_this_mediapipe_skeleton_data(skeleton3d_frame_landmark_xyz, mediapipe_landmark_names):
    number_of_frames = skeleton3d_frame_landmark_xyz.shape[0]

    number_nan_left_heel = np.sum(
        np.isnan(skeleton3d_frame_landmark_xyz[:, mediapipe_landmark_names.index("left_heel"), 0])
    )
    number_nan_right_heel = np.sum(
        np.isnan(skeleton3d_frame_landmark_xyz[:, mediapipe_landmark_names.index("right_heel"), 0])
    )
    number_nan_left_toe = np.sum(
        np.isnan(skeleton3d_frame_landmark_xyz[:, mediapipe_landmark_names.index("left_foot_index"), 0])
    )
    number_nan_right_toe = np.sum(
        np.isnan(skeleton3d_frame_landmark_xyz[:, mediapipe_landmark_names.index("right_foot_index"), 0])
    )

    no_left_heel = number_nan_left_heel == number_of_frames
    no_right_heel = number_nan_right_heel == number_of_frames
    no_left_toe = number_nan_left_toe == number_of_frames
    no_right_toe = number_nan_right_toe == number_of_frames

    if no_left_heel and no_right_heel and no_left_toe and no_right_toe:
        return False
    else:
        return True


def gap_fill_filter_origin_align_3d_data_and_then_calculate_center_of_mass(
        raw_skel3d_frame_marker_xyz: np.ndarray,
        skeleton_reprojection_error_fr_mar: np.ndarray,
        path_to_folder_where_we_will_save_this_data: [str, Path],
        # Filter the data, set the filtering options here
        sampling_rate: Union[float, int],
        cut_off: Union[float, int],
        order: Union[float, int],
        reference_frame_number: Union[float, int] = None,

):
    path_to_folder_where_we_will_save_this_data = Path(path_to_folder_where_we_will_save_this_data)

    logger.info("Gap-filling data...")
    # Interpolate the data
    processed_skel3d_frame_marker_xyz = fill_gaps_in_freemocap_data(raw_skel3d_frame_marker_xyz, )

    logger.info(
        f"Filtering data at with a {order}th order, zero-lag, low-pass Butterworth filter with a cut-off frequency of {cut_off} Hz..."
    )
    processed_skel3d_frame_marker_xyz = butterworth_filter_skeleton(
        processed_skel3d_frame_marker_xyz, cut_off, sampling_rate, order
    )

    # # pin skeleton to origin (set to mean positin of skeletonin this recording)
    # zeroed_skeleton_data = butterworth_filtered_skeleton_data.copy()
    # zeroed_skeleton_data[:, :, 0] -= np.nanmean(zeroed_skeleton_data[:, :, 0])
    # zeroed_skeleton_data[:, :, 1] -= np.nanmean(zeroed_skeleton_data[:, :, 1])
    # zeroed_skeleton_data[:, :, 2] -= np.nanmean(zeroed_skeleton_data[:, :, 2])
    #
    # # TODO - align skeleton so feet are at Z-plane and head is facing up
    # origin_aligned_freemocap_marker_data = zeroed_skeleton_data.copy()
    # logger.info("Aligning data to the origin...")

    #     reference_frame_number = find_good_clean_frame_reprojection_error_method(
    #         skeleton_3d_fr_mar_xyz=freemocap_filtered_marker_data,
    #         skeleton_reprojection_error_fr_mar=skeleton_reprojection_error_fr_mar,
    #     )

    # has_feet = are_there_feet_in_this_mediapipe_skeleton_data(
    #     butterworth_filtered_skeleton_data, mediapipe_landmark_names
    # )
    # if has_feet:
    #     reference_frame_number = find_good_frame_recursive_guess_method(
    #         butterworth_filtered_skeleton_data, mediapipe_landmark_names, 0.3
    #     )
    #
    #     logger.info("Using the foot/spine method of alignment...")
    #     origin_aligned_freemocap_marker_data = (
    #         align_skeleton_with_origin_foot_spine_method(
    #             butterworth_filtered_skeleton_data,
    #             mediapipe_landmark_names,
    #             reference_frame_number,
    #         )
    #     )
    # else:  # no feet
    #     logger.info("Using the skelly blob method of alignment...")
    #     origin_aligned_freemocap_marker_data = (
    #         align_skeleton_with_origin_mean_blob_method(
    #             butterworth_filtered_skeleton_data
    #         )
    #     )

    logger.info("Saving Origin Aligned Data")
    Path(path_to_folder_where_we_will_save_this_data).mkdir(parents=True, exist_ok=True)
    np.save(
        str(path_to_folder_where_we_will_save_this_data / "mediaPipeSkel_3d_body_hands_face.npy"),
        processed_skel3d_frame_marker_xyz,
    )

    # Calculate segment and total body COM
    anthropometric_info_dataframe = build_anthropometric_dataframe(
        segments, joint_connections, segment_COM_lengths, segment_COM_percentages
    )
    skelcoordinates_frame_segment_joint_XYZ = build_mediapipe_skeleton(
        processed_skel3d_frame_marker_xyz,
        anthropometric_info_dataframe,
        mediapipe_landmark_names,
    )
    (segment_COM_frame_dict, segment_COM_frame_imgPoint_XYZ, totalBodyCOM_frame_XYZ,) = calculate_center_of_mass(
        processed_skel3d_frame_marker_xyz,
        skelcoordinates_frame_segment_joint_XYZ,
        anthropometric_info_dataframe,
    )
    Path(path_to_folder_where_we_will_save_this_data / CENTER_OF_MASS_FOLDER_NAME).mkdir(parents=True, exist_ok=True)
    np.save(
        str(
            path_to_folder_where_we_will_save_this_data
            / CENTER_OF_MASS_FOLDER_NAME
            / SEGMENT_CENTER_OF_MASS_NPY_FILE_NAME
        ),
        segment_COM_frame_imgPoint_XYZ,
    )
    np.save(
        str(
            path_to_folder_where_we_will_save_this_data
            / CENTER_OF_MASS_FOLDER_NAME
            / TOTAL_BODY_CENTER_OF_MASS_NPY_FILE_NAME
        ),
        totalBodyCOM_frame_XYZ,
    )

    logger.info("Done with gap filling, filtering, aligning, and COM calculation")
    return processed_skel3d_frame_marker_xyz


if __name__ == "__main__":
    mediapipe_3d_frame_trackedPoint_xyz = np.load(
        r"C:\Users\jonma\freemocap_data\session_10-15-2022-09_50_10\output_data\raw_data\mediapipe_3dData_numFrames_numTrackedPoints_spatialXYZ.npy"
    )
    skeleton_reprojection_error_fr_mar = np.load(
        r"C:\Users\jonma\freemocap_data\session_10-15-2022-09_50_10\output_data\raw_data\mediapipe_3dData_numFrames_numTrackedPoints_reprojectionError.npy"
    )
    output_data_folder_path = r"C:\Users\jonma\freemocap_data\session_10-15-2022-09_50_10\output_data"
    gap_fill_filter_origin_align_3d_data_and_then_calculate_center_of_mass(
        mediapipe_3d_frame_trackedPoint_xyz,
        skeleton_reprojection_error_fr_mar=skeleton_reprojection_error_fr_mar,
        path_to_folder_where_we_will_save_this_data=output_data_folder_path,
        sampling_rate=30,
        cut_off=7,
        order=4,
        reference_frame_number=None,
    )
    print("Done :D  ")



from pathlib import Path
from typing import Union

from freemocap.tests.utilities.get_number_of_frames_of_videos_in_a_folder import (
    get_number_of_frames_of_videos_in_a_folder,
)
from freemocap.utilities.get_video_paths import get_video_paths


def test_synchronized_video_frame_counts(video_folder_path: Union[Path, str]):
    """
    Test if all the videos in this folder have precisely the same number of frames
    """
    list_of_video_paths = get_video_paths(Path(video_folder_path))

    assert len(list_of_video_paths) > 0, f"No videos found in {video_folder_path}"

    frame_count = get_number_of_frames_of_videos_in_a_folder(video_folder_path)

    assert (
        len(set(frame_count)) == 1
    ), f"Videos in {video_folder_path} have different frame counts: {frame_count}"

    return True



import logging
from pathlib import Path

import numpy as np
import pandas as pd
from rich import print

logger = logging.getLogger(__name__)

mediapipe_skeleton_segment_definitions = {
    "lower_spine": {"proximal": "hips_center", "distal": "chest_center"},
    "upper_spine": {"proximal": "chest_center", "distal": "neck_center"},
    "head": {"proximal": "neck_center", "distal": "head_center"},
    "left_clavicle": {"proximal": "neck_center", "distal": "left_shoulder"},
    "left_upper_arm": {"proximal": "left_shoulder", "distal": "left_elbow"},
    "left_forearm": {"proximal": "left_elbow", "distal": "left_wrist"},
    "left_hand": {"proximal": "left_wrist", "distal": "left_index"},
    "right_clavicle": {"proximal": "neck_center", "distal": "right_shoulder"},
    "right_upper_arm": {"proximal": "right_shoulder", "distal": "right_elbow"},
    "right_forearm": {"proximal": "right_elbow", "distal": "right_wrist"},
    "right_hand": {"proximal": "right_wrist", "distal": "right_index"},
    "left_pelvis": {"proximal": "hips_center", "distal": "left_hip"},
    "left_thigh": {"proximal": "left_hip", "distal": "left_knee"},
    "left_calf": {"proximal": "left_knee", "distal": "left_ankle"},
    "left_foot": {"proximal": "left_ankle", "distal": "left_foot_index"},
    "right_pelvis": {"proximal": "hips_center", "distal": "right_hip"},
    "right_thigh": {"proximal": "right_hip", "distal": "right_knee"},
    "right_calf": {"proximal": "right_knee", "distal": "right_ankle"},
    "right_foot": {"proximal": "right_ankle", "distal": "right_foot_index"},
}


def estimate_skeleton_segment_lengths(skeleton_dataframe: pd.DataFrame, skeleton_segment_definitions: dict) -> dict:
    """Estimate the length of each skeleton segment.

    Args:
        skeleton_data_dictionary (dict): Dictionary containing the skeleton data.
        skeleton_segment_definitions (dict): Dictionary containing the definitions of each segment (i.e. the proximal and distal joints).

    Returns:
        dict: Dictionary containing the estimated length of each skeleton segment.
    """
    logger.debug("Estimating skeleton segment lengths (mm)...")
    skeleton_data_dictionary = create_skeleton_dictionary_from_skeleton_body_data_frame(skeleton_dataframe)

    skeleton_segment_lengths_dict = {}
    for segment_name, segment_definition_dict in skeleton_segment_definitions.items():
        print(
            f"Estimating length of segment '{segment_name}' "
            f"based on proximal joint: '{segment_definition_dict['proximal']}' and "
            f"distal joint:'{segment_definition_dict['distal']}'"
        )
        proximal_joint_name = segment_definition_dict["proximal"]
        distal_joint_name = segment_definition_dict["distal"]

        proximal_x = skeleton_data_dictionary[f"{proximal_joint_name}_x"]
        proximal_y = skeleton_data_dictionary[f"{proximal_joint_name}_y"]
        proximal_z = skeleton_data_dictionary[f"{proximal_joint_name}_z"]

        distal_x = skeleton_data_dictionary[f"{distal_joint_name}_x"]
        distal_y = skeleton_data_dictionary[f"{distal_joint_name}_y"]
        distal_z = skeleton_data_dictionary[f"{distal_joint_name}_z"]

        # logger.debug(f"proximal_x nans: {np.isnan(proximal_x).sum()}")
        # logger.debug(f"len proximal_x: {len(proximal_x)}")
        # logger.debug(f"distal_x nans: {np.isnan(distal_x).sum()}")
        # logger.debug(f"len distal_x: {len(distal_x)}")

        # good ol pythag <3
        segment_length_per_frame = np.sqrt(
            (proximal_x - distal_x) ** 2 + (proximal_y - distal_y) ** 2 + (proximal_z - distal_z) ** 2
        )
        skeleton_segment_lengths_dict[segment_name] = {}
        skeleton_segment_lengths_dict[segment_name]["median"] = np.nanmedian(segment_length_per_frame)
        skeleton_segment_lengths_dict[segment_name]["mean"] = np.nanmean(segment_length_per_frame)
        skeleton_segment_lengths_dict[segment_name]["standard_deviation"] = np.nanstd(segment_length_per_frame)

    return skeleton_segment_lengths_dict


def create_virtual_markers(skeleton_data_dictionary, virtual_marker_name: str, marker_name_list: list) -> dict:
    # print(
    #     f"Creating virtual marker '{virtual_marker_name}' from markers: {marker_name_list}"
    # )

    x = []
    y = []
    z = []

    for marker_name in marker_name_list:
        x.append(skeleton_data_dictionary[f"{marker_name}_x"])
        y.append(skeleton_data_dictionary[f"{marker_name}_y"])
        z.append(skeleton_data_dictionary[f"{marker_name}_z"])

    skeleton_data_dictionary[f"{virtual_marker_name}_x"] = np.nanmean(np.asarray(x), axis=0)
    skeleton_data_dictionary[f"{virtual_marker_name}_y"] = np.nanmean(np.asarray(y), axis=0)
    skeleton_data_dictionary[f"{virtual_marker_name}_z"] = np.nanmean(np.asarray(z), axis=0)

    return skeleton_data_dictionary


def create_skeleton_dictionary_from_skeleton_body_data_frame(
    skeleton_body_dataframe: pd.DataFrame,
) -> dict:
    skeleton_data_dictionary = skeleton_body_dataframe.to_dict(orient="list")

    for key, value in skeleton_data_dictionary.items():
        skeleton_data_dictionary[key] = np.asarray(value)

    skeleton_data_dictionary = create_virtual_markers(
        skeleton_data_dictionary, "hips_center", ["left_hip", "right_hip"]
    )
    skeleton_data_dictionary = create_virtual_markers(
        skeleton_data_dictionary, "neck_center", ["left_shoulder", "right_shoulder"]
    )
    skeleton_data_dictionary = create_virtual_markers(
        skeleton_data_dictionary, "head_center", ["left_ear", "right_ear"]
    )

    skeleton_data_dictionary = create_virtual_markers(
        skeleton_data_dictionary, "chest_center", ["hips_center", "neck_center"]
    )
    return skeleton_data_dictionary


if __name__ == "__main__":
    path_to_skeleton_body_csv = Path(
        r"D:\Dropbox\FreeMoCapProject\teddy_animation\FreeMocap_Data\sesh_2022-10-09_13_55_45_calib_and_take_1_alpha\output_data\mediapipe_body_3d_xyz.csv"
    )
    skeleton_dataframe = pd.read_csv(path_to_skeleton_body_csv)

    skeleton_segment_lengths = estimate_skeleton_segment_lengths(
        skeleton_dataframe, mediapipe_skeleton_segment_definitions
    )

    print(skeleton_segment_lengths)



import logging

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QTabWidget, QVBoxLayout, QWidget
# from skelly_viewer import SkellyViewer
from skellycam import SkellyCamWidget

from freemocap.gui.qt.widgets.home_widget import (
    HomeWidget,
)
from freemocap.system.paths_and_files_names import CAMERA_WITH_FLASH_EMOJI_STRING, EYES_EMOJI_STRING

logger = logging.getLogger(__name__)


class CentralTabWidget(QTabWidget):
    def __init__(
        self,
        skelly_cam_widget: SkellyCamWidget,
        camera_controller_widget: QWidget,
        welcome_to_freemocap_widget: HomeWidget,
        skelly_viewer_widget: QWidget,
        directory_view_widget: QWidget,
        active_recording_info_widget: QWidget,
        parent=None,
    ):
        super().__init__(parent=parent)
        self.parent = parent

        # self.setTabShape(QTabWidget.TabShape.Triangular)

        self._skelly_cam_widget = skelly_cam_widget
        self._camera_controller_widget = camera_controller_widget
        self._welcome_to_freemocap_widget = welcome_to_freemocap_widget
        self._skelly_viewer_widget = skelly_viewer_widget
        self._directory_view_widget = directory_view_widget
        self._active_recording_info_widget = active_recording_info_widget

        self._create_welcome_tab(self)
        self._create_skellycam_view_tab(self)
        self._create_skelly_viewer_tab(self)
        self._create_directory_view_tab(self)
        self._create_active_recording_info_tab(self)

    def set_welcome_tab_enabled(self, enabled: bool):
        self.setTabEnabled(0, enabled)

    def set_camera_view_tab_enabled(self, enabled: bool):
        self.setTabEnabled(1, enabled)

    def set_visualize_data_tab_enabled(self, enabled: bool):
        self.setTabEnabled(2, enabled)

    def _create_welcome_tab(self, tab_widget: QTabWidget):
        logger.info("Creating welcome tab")
        tab_widget.addTab(self._welcome_to_freemocap_widget, "Home")

    def _create_skellycam_view_tab(self, tab_widget: QTabWidget):
        logger.info("Creating skellycam view tab")
        dummy_widget = QWidget()
        self._camera_view_layout = QVBoxLayout()
        self._camera_view_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        dummy_widget.setLayout(self._camera_view_layout)
        tab_widget.addTab(dummy_widget, f"Cameras")
        # tab_widget.setToolTip(skellycam.__repo_url__)

        # self._qt_multi_camera_viewer_widget.resize(1280, 720)
        self._camera_view_layout.addWidget(self._camera_controller_widget)

        self._camera_view_layout.addWidget(self._skelly_cam_widget)

    def _create_skelly_viewer_tab(self, tab_widget: QTabWidget):
        logger.info("Creating export_data tab")
        tab_widget.addTab(self._skelly_viewer_widget, f"Data Viewer")
        # tab_widget.setToolTip(skelly_viewer.__repo_url__)

    def _create_directory_view_tab(self, tab_widget: QTabWidget):
        logger.info("Creating directory view tab")
        tab_widget.addTab(self._directory_view_widget, f"Directory View")

    def _create_active_recording_info_tab(self, tab_widget: QTabWidget):
        logger.info("Creating active recording info tab")
        tab_widget.addTab(self._active_recording_info_widget, f"Active Recording Info")




import logging

logger = logging.getLogger(__name__)

if __name__ == "__main__":
    logger.info(
        "User tried using `pre-alpha` entry point (`import freemocap: freemocap.RunMe() - displaying friendly message then re-directing to `freemocap.__main__:main()` entry point")

    print("--------------------------------\n"
          "--------------------------------\n"
          "--------------------------------\n"
          "Hello! Looks like you're trying to use the `alpha GUI` entry point for FreeMoCap.\n"
          "This entry point is deprecated, so we're launching the GUI via `freemocap.__main__:main()` entry point.\n"
          "If you want use the `alpha GUI` code, use check out the `v0.1.0` tag in the github repo.\n"
          "Thank you for using FreeMoCap!\n"
          "--------------------------------\n"
          "(NOTE  - this entry point will be removed eventually\n"
          "--------------------------------\n"
          "--------------------------------\n")

    from freemocap.__main__ import main

    main()



import logging
from pathlib import Path
from typing import Union, Dict, Any

import numpy as np
import toml

from freemocap.system.paths_and_files_names import (
    CENTER_OF_MASS_FOLDER_NAME,
    MEDIAPIPE_2D_NPY_FILE_NAME,
    RAW_MEDIAPIPE_3D_NPY_FILE_NAME,
    OUTPUT_DATA_FOLDER_NAME,
    RAW_DATA_FOLDER_NAME,
    TOTAL_BODY_CENTER_OF_MASS_NPY_FILE_NAME,
    MEDIAPIPE_REPROJECTION_ERROR_NPY_FILE_NAME,
    SYNCHRONIZED_VIDEOS_FOLDER_NAME,
    get_blender_file_path,
    ANNOTATED_VIDEOS_FOLDER_NAME,
    MEDIAPIPE_3D_NPY_FILE_NAME, create_camera_calibration_file_name,
)
from freemocap.tests.test_mediapipe_2d_data_shape import test_mediapipe_2d_data_shape
from freemocap.tests.test_mediapipe_3d_data_shape import test_mediapipe_3d_data_shape
from freemocap.tests.test_synchronized_video_frame_counts import (
    test_synchronized_video_frame_counts,
)
from freemocap.tests.test_total_body_center_of_mass_data_shape import (
    test_total_body_center_of_mass_data_shape,
)

logger = logging.getLogger(__name__)


class RecordingInfoModel:
    def __init__(
            self,
            recording_folder_path: Union[Path, str],
    ):
        if any([Path(recording_folder_path).name == SYNCHRONIZED_VIDEOS_FOLDER_NAME,
                Path(recording_folder_path).name == ANNOTATED_VIDEOS_FOLDER_NAME,
                Path(recording_folder_path).name == OUTPUT_DATA_FOLDER_NAME]):
            recording_folder_path = Path(recording_folder_path).parent

        self._path = Path(recording_folder_path)
        self._name = self._path.name

        self._calibration_toml_path = str(
            Path(self._path) / create_camera_calibration_file_name(recording_name=self._name))

        self._recording_folder_status_checker = RecordingFolderStatusChecker(recording_info_model=self)

    @property
    def path(self) -> str:
        return str(self._path)

    @property
    def name(self) -> str:
        return self._name

    @property
    def status_check(self) -> Dict[str, bool]:
        return self._recording_folder_status_checker.status_check

    @property
    def calibration_toml_path(self) -> str:
        return self._calibration_toml_path
    @calibration_toml_path.setter
    def calibration_toml_path(self, path: Union[Path, str]):
        self._calibration_toml_path = str(path)

    @property
    def output_data_folder_path(self) -> str:
        return str(Path(self._path) / OUTPUT_DATA_FOLDER_NAME)

    @property
    def raw_data_folder_path(self) -> str:
        return str(Path(self.output_data_folder_path) / RAW_DATA_FOLDER_NAME)

    @property
    def synchronized_videos_folder_path(self) -> str:
        return str(Path(self._path) / SYNCHRONIZED_VIDEOS_FOLDER_NAME)

    @property
    def annotated_videos_folder_path(self) -> str:
        return str(Path(self._path) / ANNOTATED_VIDEOS_FOLDER_NAME)

    @property
    def mediapipe_2d_data_npy_file_path(self):
        return str(Path(self._path) / OUTPUT_DATA_FOLDER_NAME / RAW_DATA_FOLDER_NAME / MEDIAPIPE_2D_NPY_FILE_NAME)

    @property
    def mediapipe_3d_data_npy_file_path(self):
        return str(Path(self._path) / OUTPUT_DATA_FOLDER_NAME / MEDIAPIPE_3D_NPY_FILE_NAME)

    @property
    def raw_mediapipe_3d_data_npy_file_path(self):
        return str(Path(self._path) / OUTPUT_DATA_FOLDER_NAME / RAW_DATA_FOLDER_NAME / RAW_MEDIAPIPE_3D_NPY_FILE_NAME)

    @property
    def mediapipe_reprojection_error_data_npy_file_path(self):
        return str(
            Path(self._path)
            / OUTPUT_DATA_FOLDER_NAME
            / RAW_DATA_FOLDER_NAME
            / MEDIAPIPE_REPROJECTION_ERROR_NPY_FILE_NAME
        )

    @property
    def total_body_center_of_mass_npy_file_path(self):
        return str(
            Path(self._path)
            / OUTPUT_DATA_FOLDER_NAME
            / CENTER_OF_MASS_FOLDER_NAME
            / TOTAL_BODY_CENTER_OF_MASS_NPY_FILE_NAME
        )

    @property
    def blender_file_path(self):
        return get_blender_file_path(str(self._path))

    @property
    def blender_file_status_check(self):
        return Path(self.blender_file_path).is_file()

    @property
    def calibration_toml_check(self) -> bool:
        return self._recording_folder_status_checker.check_calibration_toml_status()

    @property
    def synchronized_videos_status_check(self) -> bool:
        return self._recording_folder_status_checker.check_synchronized_videos_status()

    @property
    def data2d_status_check(self) -> bool:
        return self._recording_folder_status_checker.check_data2d_status()

    @property
    def data3d_status_check(self) -> bool:
        return self._recording_folder_status_checker.check_data3d_status()

    @property
    def center_of_mass_data_status_check(self) -> bool:
        return self._recording_folder_status_checker.check_center_of_mass_data_status()


class RecordingFolderStatusChecker:
    def __init__(self, recording_info_model: RecordingInfoModel):

        self.recording_info_model = recording_info_model

    @property
    def status_check(self) -> Dict[str, Union[bool, str, float]]:
        return {
            'synchronized_videos_status_check': self.check_synchronized_videos_status(),
            'data2d_status_check': self.check_data2d_status(),
            'data3d_status_check': self.check_data3d_status(),
            'center_of_mass_data_status_check': self.check_center_of_mass_data_status(),
            'blender_file_status_check': self.check_blender_file_status(),
            'video_and_camera_info': {
                'number_of_synchronized_videos': self.get_number_of_mp4s_in_synched_videos_directory(),
                'number_of_frames_in_videos': self.get_number_of_frames_in_videos(),
                # 'camera_rotation_and_translation': self.get_camera_rotation_and_translation_from_calibration_toml()
            }
        }

    def check_synchronized_videos_status(self) -> bool:
        try:
            test_synchronized_video_frame_counts(self.recording_info_model.synchronized_videos_folder_path)
            return True
        except AssertionError:
            return False

    def check_data2d_status(self) -> bool:

        try:
            test_mediapipe_2d_data_shape(
                synchronized_videos_folder=self.recording_info_model.synchronized_videos_folder_path,
                mediapipe_2d_data_file_path=self.recording_info_model.mediapipe_2d_data_npy_file_path,
            )

            return True
        except AssertionError as e:
            return False

    def check_data3d_status(self) -> bool:
        try:
            test_mediapipe_3d_data_shape(
                synchronized_videos_folder=self.recording_info_model.synchronized_videos_folder_path,
                mediapipe_3d_data_npy_path=self.recording_info_model.mediapipe_3d_data_npy_file_path,
                medipipe_reprojection_error_data_npy_path=self.recording_info_model.mediapipe_reprojection_error_data_npy_file_path,
            )
            return True
        except AssertionError as e:

            return False

    def check_center_of_mass_data_status(self) -> bool:
        try:
            test_total_body_center_of_mass_data_shape(
                synchronized_videos_folder=self.recording_info_model.synchronized_videos_folder_path,
                total_body_center_of_mass_npy_file_path=self.recording_info_model.total_body_center_of_mass_npy_file_path,
            )
            return True
        except AssertionError as e:
            return False

    def check_blender_file_status(self) -> bool:
        return Path(self.recording_info_model.blender_file_path).is_file()

    def check_calibration_toml_status(self) -> bool:
        return Path(self.recording_info_model.calibration_toml_path).is_file()

    def get_number_of_mp4s_in_synched_videos_directory(self) -> float:

        synchronized_directory_path = Path(self.recording_info_model.synchronized_videos_folder_path)

        if not synchronized_directory_path.exists():
            return 0

        video_count = 0
        for file in synchronized_directory_path.iterdir():
            if file.is_file() and file.suffix.lower() == '.mp4':
                video_count += 1
        return video_count

    def get_number_of_frames_in_videos(self):
        timestamps_directory_path = Path(self.recording_info_model.synchronized_videos_folder_path) / "timestamps"

        if timestamps_directory_path.exists() and timestamps_directory_path.is_dir():
            frame_counts = {}

            for npy_file in timestamps_directory_path.iterdir():
                if npy_file.is_file() and npy_file.suffix.lower() == '.npy':
                    video_npy = np.load(str(npy_file))
                    frame_counts[npy_file.name] = str(len(video_npy)-1)

            return frame_counts
        else:
            return "No 'timestamps' directory found"

    # def get_camera_rotation_and_translation_from_calibration_toml(self):
    #     calibration_toml_path = Path(self.recording_info_model.calibration_toml_path)
    #
    #     logger.info(f"Looking for calibration file at {calibration_toml_path}")
    #
    #     if calibration_toml_path.exists() and calibration_toml_path.is_file():
    #         calibration_data = toml.load(str(calibration_toml_path))
    #
    #         camera_data = {}
    #
    #         for cam_key, cam_values in calibration_data.items():
    #             if cam_key.startswith("cam_"):
    #                 camera_data[cam_key] = {
    #                     'rotation': cam_values['rotation'],
    #                     'translation': cam_values['translation']
    #                 }
    #
    #         return camera_data
    #     else:
    #         return "No 'calibration.toml' file found"



import logging
import multiprocessing
from pathlib import Path

import pandas as pd

from freemocap.core_processes.detecting_things_in_2d_images.mediapipe_stuff.mediapipe_skeleton_names_and_connections import (
    mediapipe_names_and_connections_dict,
)
from freemocap.core_processes.post_process_skeleton_data.process_single_camera_skeleton_data import \
    process_single_camera_skeleton_data
from freemocap.system.paths_and_files_names import (
    MEDIAPIPE_BODY_3D_DATAFRAME_CSV_FILE_NAME,
    RAW_DATA_FOLDER_NAME,
)
from freemocap.core_processes.capture_volume_calibration.anipose_camera_calibration.get_anipose_calibration_object import (
    load_anipose_calibration_toml_from_path,
)
from freemocap.core_processes.capture_volume_calibration.triangulate_3d_data import (
    triangulate_3d_data,
)
from freemocap.core_processes.detecting_things_in_2d_images.mediapipe_stuff.convert_mediapipe_npy_to_csv import (
    convert_mediapipe_npy_to_csv,
)
from freemocap.core_processes.detecting_things_in_2d_images.mediapipe_stuff.mediapipe_skeleton_detector import (
    MediaPipeSkeletonDetector,
)
from freemocap.core_processes.post_process_skeleton_data.estimate_skeleton_segment_lengths import (
    estimate_skeleton_segment_lengths,
    mediapipe_skeleton_segment_definitions,
)
from freemocap.core_processes.post_process_skeleton_data.gap_fill_filter_and_origin_align_skeleton_data import (
    gap_fill_filter_origin_align_3d_data_and_then_calculate_center_of_mass,
)
from freemocap.parameter_info_models.recording_processing_parameter_models import RecordingProcessingParameterModel

from freemocap.tests.test_mediapipe_2d_data_shape import (
    test_mediapipe_2d_data_shape,
)
from freemocap.tests.test_mediapipe_3d_data_shape import test_mediapipe_3d_data_shape
from freemocap.utilities.rotate_by_90_degrees_around_x_axis import rotate_by_90_degrees_around_x_axis
from freemocap.utilities.save_dictionary_to_json import save_dictionary_to_json

logger = logging.getLogger(__name__)


def process_recording_folder(
        recording_processing_parameter_model: RecordingProcessingParameterModel,
        kill_event: multiprocessing.Event = None,
):
    """

    Parameters
    ----------
    recording_processing_parameter_model : RecordingProcessingParameterModel
        RecordingProcessingParameterModel object (contains all the paths and parameters necessary to process a session folder

    """

    rec = recording_processing_parameter_model  # make it smol

    if not Path(rec.recording_info_model.synchronized_videos_folder_path).exists():
        raise FileNotFoundError(
            f"Could not find synchronized_videos folder at {rec.recording_info_model.synchronized_videos_folder_path}"
        )

    logger.info("Detecting 2d skeletons...")
    # 2d skeleton detection
    mediapipe_skeleton_detector = MediaPipeSkeletonDetector(
        parameter_model=rec.mediapipe_parameters_model,
    )

    mediapipe_image_data_numCams_numFrames_numTrackedPts_XYZ = mediapipe_skeleton_detector.process_folder_full_of_videos(
        rec.recording_info_model.synchronized_videos_folder_path,
        Path(rec.recording_info_model.output_data_folder_path) / RAW_DATA_FOLDER_NAME,
        kill_event=kill_event,
    )

    if kill_event is not None and kill_event.is_set():
        return

    assert test_mediapipe_2d_data_shape(
        synchronized_videos_folder=rec.recording_info_model.synchronized_videos_folder_path,
        mediapipe_2d_data_file_path=rec.recording_info_model.mediapipe_2d_data_npy_file_path,
    )

    # spoof 3D data if single camera
    if mediapipe_image_data_numCams_numFrames_numTrackedPts_XYZ.shape[0] == 1:

        (raw_skel3d_frame_marker_xyz, skeleton_reprojection_error_fr_mar) = process_single_camera_skeleton_data(
            input_image_data_frame_marker_xyz=mediapipe_image_data_numCams_numFrames_numTrackedPts_XYZ[0],
            raw_data_folder_path=Path(rec.recording_info_model.raw_data_folder_path))

    else:

        logger.info("Triangulating 3d skeletons...")

        assert rec.recording_info_model.calibration_toml_check, f"No calibration file found at: {rec.recording_info_model.calibration_toml_path}"
        assert rec.recording_info_model.data2d_status_check, f"No mediapipe 2d data found at: {rec.recording_info_model.mediapipe_2d_data_npy_file_path}"

        anipose_calibration_object = load_anipose_calibration_toml_from_path(
            camera_calibration_data_toml_path=rec.recording_info_model.calibration_toml_path,
            save_copy_of_calibration_to_this_path=rec.recording_info_model.path,
        )
        (raw_skel3d_frame_marker_xyz, skeleton_reprojection_error_fr_mar,) = triangulate_3d_data(
            anipose_calibration_object=anipose_calibration_object,
            mediapipe_2d_data=mediapipe_image_data_numCams_numFrames_numTrackedPts_XYZ[:, :, :, :2],
            output_data_folder_path=rec.recording_info_model.raw_data_folder_path,
            mediapipe_confidence_cutoff_threshold=rec.anipose_triangulate_3d_parameters_model.confidence_threshold_cutoff,
            use_triangulate_ransac=rec.anipose_triangulate_3d_parameters_model.use_triangulate_ransac_method,
            kill_event=kill_event,
        )

        if kill_event is not None and kill_event.is_set():
            return

        assert test_mediapipe_3d_data_shape(
            synchronized_videos_folder=rec.recording_info_model.synchronized_videos_folder_path,
            mediapipe_3d_data_npy_path=rec.recording_info_model.raw_mediapipe_3d_data_npy_file_path,
            medipipe_reprojection_error_data_npy_path=rec.recording_info_model.mediapipe_reprojection_error_data_npy_file_path,
        )

    #rotate so skeleton is closer to 'vertical' in a z-up reference frame
    rotated_raw_skel3d_frame_marker_xyz = rotate_by_90_degrees_around_x_axis(raw_skel3d_frame_marker_xyz) 

    logger.info("Gap-filling, butterworth filtering, origin aligning 3d skeletons, then calculating center of mass ...")

    skel3d_frame_marker_xyz = gap_fill_filter_origin_align_3d_data_and_then_calculate_center_of_mass(
        raw_skel3d_frame_marker_xyz=rotated_raw_skel3d_frame_marker_xyz,
        skeleton_reprojection_error_fr_mar=skeleton_reprojection_error_fr_mar,
        path_to_folder_where_we_will_save_this_data=rec.recording_info_model.output_data_folder_path,
        sampling_rate=rec.post_processing_parameters_model.framerate,
        cut_off=rec.post_processing_parameters_model.butterworth_filter_parameters.cutoff_frequency,
        order=rec.post_processing_parameters_model.butterworth_filter_parameters.order,
        reference_frame_number=None,
    )
    assert test_mediapipe_3d_data_shape(
        synchronized_videos_folder=rec.recording_info_model.synchronized_videos_folder_path,
        mediapipe_3d_data_npy_path=rec.recording_info_model.mediapipe_3d_data_npy_file_path,
        medipipe_reprojection_error_data_npy_path=rec.recording_info_model.mediapipe_reprojection_error_data_npy_file_path,
    )

    logger.info("Breaking up big `npy` into smaller bits and converting to `csv`...")
    # break up big NPY and save out csv's
    convert_mediapipe_npy_to_csv(
        mediapipe_3d_frame_trackedPoint_xyz=skel3d_frame_marker_xyz,
        output_data_folder_path=rec.recording_info_model.output_data_folder_path,
    )

    path_to_skeleton_body_csv = (
            Path(rec.recording_info_model.output_data_folder_path) / MEDIAPIPE_BODY_3D_DATAFRAME_CSV_FILE_NAME
    )
    skeleton_dataframe = pd.read_csv(path_to_skeleton_body_csv)

    logger.info("Estimating skeleton segment lengths...")
    skeleton_segment_lengths_dict = estimate_skeleton_segment_lengths(
        skeleton_dataframe=skeleton_dataframe,
        skeleton_segment_definitions=mediapipe_skeleton_segment_definitions,
    )

    save_dictionary_to_json(
        save_path=rec.recording_info_model.output_data_folder_path,
        file_name="mediapipe_skeleton_segment_lengths.json",
        dictionary=skeleton_segment_lengths_dict,
    )

    save_dictionary_to_json(
        save_path=rec.recording_info_model.output_data_folder_path,
        file_name="mediapipe_names_and_connections_dict.json",
        dictionary=mediapipe_names_and_connections_dict,
    )



import logging
import time
from datetime import datetime
from pathlib import Path
from typing import Union

import toml

import freemocap

logger = logging.getLogger(__name__)

# directory names
BASE_FREEMOCAP_DATA_FOLDER_NAME = "freemocap_data"
RECORDING_SESSIONS_FOLDER_NAME = "recording_sessions"
CALIBRATIONS_FOLDER_NAME = "calibrations"
LOGS_INFO_AND_SETTINGS_FOLDER_NAME = "logs_info_and_settings"
LOG_FILE_FOLDER_NAME = "logs"
OUTPUT_DATA_FOLDER_NAME = "output_data"
SYNCHRONIZED_VIDEOS_FOLDER_NAME = "synchronized_videos"
ANNOTATED_VIDEOS_FOLDER_NAME = "annotated_videos"
RAW_DATA_FOLDER_NAME = "raw_data"
CENTER_OF_MASS_FOLDER_NAME = "center_of_mass"

# file names
MOST_RECENT_RECORDING_TOML_FILENAME = "most_recent_recording.toml"
LAST_SUCCESSFUL_CALIBRATION_TOML_FILENAME = "last_successful_calibration.toml"
MEDIAPIPE_2D_NPY_FILE_NAME = "mediapipe2dData_numCams_numFrames_numTrackedPoints_pixelXY.npy"
MEDIAPIPE_BODY_WORLD_FILE_NAME = "mediapipeBodyWorld_numCams_numFrames_numTrackedPoints_XYZ.npy"
MEDIAPIPE_3D_NPY_FILE_NAME = "mediaPipeSkel_3d_body_hands_face.npy"
RAW_MEDIAPIPE_3D_NPY_FILE_NAME = "mediapipe3dData_numFrames_numTrackedPoints_spatialXYZ.npy"
MEDIAPIPE_REPROJECTION_ERROR_NPY_FILE_NAME = "mediapipe3dData_numFrames_numTrackedPoints_reprojectionError.npy"
MEDIAPIPE_BODY_3D_DATAFRAME_CSV_FILE_NAME = "mediapipe_body_3d_xyz.csv"

SEGMENT_CENTER_OF_MASS_NPY_FILE_NAME = "segmentCOM_frame_joint_xyz.npy"

TOTAL_BODY_CENTER_OF_MASS_NPY_FILE_NAME = "total_body_center_of_mass_xyz.npy"

# logo
PATH_TO_FREEMOCAP_LOGO_SVG = str(Path(freemocap.__file__).parent.parent / "assets/logo/freemocap-logo-black-border.svg")

# emoji strings
SPARKLES_EMOJI_STRING = "\U00002728"
DIZZY_EMOJI_STRING = "\U0001F4AB"
EYES_EMOJI_STRING = "\U0001F440"
SKULL_EMOJI_STRING = "\U0001F480"
CAMERA_WITH_FLASH_EMOJI_STRING = "\U0001F4F8"
HAMMER_AND_WRENCH_EMOJI_STRING = "\U0001F6E0"
ROBOT_EMOJI_STRING = "\U0001F916"
THREE_HEARTS_EMOJI_STRING = "\U0001F495"
WIND_EMOJI_STRING = "\U0001F32C"
GEAR_EMOJI_STRING = "\U00002699"
FOLDER_EMOJI_STRING = "\U0001F4C1"
DIRECTORY_EMOJI_STRING = "\U0001F4C1"
COOL_EMOJI_STRING = "\U0001F60E"


def os_independent_home_dir():
    return str(Path.home())


def get_log_file_path():
    log_folder_path = Path(get_freemocap_data_folder_path()) / LOGS_INFO_AND_SETTINGS_FOLDER_NAME / LOG_FILE_FOLDER_NAME
    log_folder_path.mkdir(exist_ok=True, parents=True)
    log_file_path = log_folder_path / create_log_file_name()
    return str(log_file_path)


def create_log_file_name():
    return "log_" + time.strftime("%m-%d-%Y-%H_%M_%S") + ".log"


def create_camera_calibration_file_name(recording_name: str):
    return f"{recording_name}_camera_calibration.toml"


freemocap_data_folder_path = None
def get_freemocap_data_folder_path(create_folder: bool = True):
    global freemocap_data_folder_path

    if freemocap_data_folder_path is None:
        freemocap_data_folder_path = Path(os_independent_home_dir(), BASE_FREEMOCAP_DATA_FOLDER_NAME)

        if create_folder:
            freemocap_data_folder_path.mkdir(exist_ok=create_folder, parents=True)

    return str(freemocap_data_folder_path)


def get_calibrations_folder_path(create_folder: bool = True):
    calibration_folder_path = Path(get_freemocap_data_folder_path()) / CALIBRATIONS_FOLDER_NAME

    if create_folder:
        calibration_folder_path.mkdir(exist_ok=create_folder, parents=True)

    return str(calibration_folder_path)


def create_new_recording_folder_path(recording_name: str):

    recording_folder_path = Path(create_new_session_folder()) / recording_name

    return str(recording_folder_path)


def get_gmt_offset_string():
    # from - https://stackoverflow.com/a/53860920/14662833
    gmt_offset_int = int(time.localtime().tm_gmtoff / 60 / 60)
    return f"{gmt_offset_int:+}"


def get_iso6201_time_string(timespec: str = "milliseconds", make_filename_friendly: bool = True):
    iso6201_timestamp = datetime.now().isoformat(timespec=timespec)
    gmt_offset_string = f"_gmt{get_gmt_offset_string()}"
    iso6201_timestamp_w_gmt = iso6201_timestamp + gmt_offset_string
    if make_filename_friendly:
        iso6201_timestamp_w_gmt = iso6201_timestamp_w_gmt.replace(":", "_")
        iso6201_timestamp_w_gmt = iso6201_timestamp_w_gmt.replace(".", "ms")
    return iso6201_timestamp_w_gmt


def create_new_default_recording_name():
    full_time = get_iso6201_time_string(timespec="seconds")
    just_hours_minutes_seconds = full_time.split("T")[1]
    recording_name = "recording_" + just_hours_minutes_seconds
    return recording_name


def session_time_tag_format():
    return "%Y-%m-%d_%H_%M_%S"
def default_session_name(string_tag: str = None):
    if string_tag is not None:
        string_tag = f"_{string_tag}"
    else:
        string_tag = ""

    session_name = time.strftime(f"session_{session_time_tag_format()}" + string_tag)
    logger.debug(f"Creating default session name: {session_name}")

    return session_name


session_folder_path = None
def create_new_session_folder():
    global session_folder_path
    if session_folder_path is None:
        session_folder_path = Path(get_recording_session_folder_path()) / default_session_name()

    return str(session_folder_path)


def get_recording_session_folder_path(create_folder: bool = True):
    recording_session_folder_path = Path(get_freemocap_data_folder_path()) / RECORDING_SESSIONS_FOLDER_NAME

    if create_folder:
        recording_session_folder_path.mkdir(exist_ok=create_folder, parents=True)

    return str(recording_session_folder_path)


def get_logs_info_and_settings_folder_path(create_folder: bool = True):
    logs_info_and_settings_folder_path = Path(get_freemocap_data_folder_path()) / LOGS_INFO_AND_SETTINGS_FOLDER_NAME

    if create_folder:
        logs_info_and_settings_folder_path.mkdir(exist_ok=create_folder, parents=True)

    return str(logs_info_and_settings_folder_path)


def get_css_stylesheet_path():
    # return str(Path(__file__).parent.parent / "gui" / "qt" / "style_sheet" / "ElegantDark.qss")
    # return str(Path(__file__).parent.parent / "gui" / "qt" / "style_sheet" / "FunkyTown.qss")
    return str(Path(__file__).parent.parent / "gui" / "qt" / "style_sheet" / "qt_style_sheet.css")


def get_scss_stylesheet_path():
    return str(Path(__file__).parent.parent / "gui" / "qt" / "style_sheet" / "qt_style_sheet.scss")

def get_most_recent_recording_toml_path():
    return str(Path(get_logs_info_and_settings_folder_path()) / MOST_RECENT_RECORDING_TOML_FILENAME)


def get_most_recent_recording_path(subfolder_str: str = None):
    if not Path(get_most_recent_recording_toml_path()).exists():
        logger.error(f"{MOST_RECENT_RECORDING_TOML_FILENAME} not found at {get_most_recent_recording_toml_path()}!!")
        return None

    most_recent_recording_dict = toml.load(str(get_most_recent_recording_toml_path()))
    most_recent_recording_path = most_recent_recording_dict["most_recent_recording_path"]

    if not Path(most_recent_recording_path).exists():
        logger.error(f"Most recent recording path {most_recent_recording_path} not found!!")
        return None

    if subfolder_str is not None:
        try:
            # check if folder is specified in toml file, otherwise return default
            sub_folder_path = most_recent_recording_dict[subfolder_str]
        except KeyError:
            sub_folder_path = Path(most_recent_recording_path) / subfolder_str
        return str(sub_folder_path)

    else:

        return str(most_recent_recording_path)


def get_last_successful_calibration_toml_path():
    return str(Path(get_logs_info_and_settings_folder_path()) / LAST_SUCCESSFUL_CALIBRATION_TOML_FILENAME)


def get_last_successful_calibration_name():
    path = get_last_successful_calibration_toml_path()
    if not Path(path).exists():
        logger.error(f"{LAST_SUCCESSFUL_CALIBRATION_TOML_FILENAME} not found at {path}!!")
        return None
    toml_dict = toml.load(path)


def get_blender_file_name(recording_name: str):
    return f"{recording_name}.blend"


def get_blender_file_path(recording_folder_path: Union[Path,str]):
    return str(Path(recording_folder_path) / get_blender_file_name(recording_name=Path(recording_folder_path).name))



import logging
import multiprocessing
from pathlib import Path
from typing import Union

import numpy as np

from freemocap.system.paths_and_files_names import (
    RAW_MEDIAPIPE_3D_NPY_FILE_NAME,
    MEDIAPIPE_REPROJECTION_ERROR_NPY_FILE_NAME,
)

logger = logging.getLogger(__name__)


def threshold_by_confidence(
    mediapipe_2d_data: np.ndarray,
    mediapipe_confidence_cutoff_threshold: float = 0.0,
):
    mediapipe_2d_data[mediapipe_2d_data <= mediapipe_confidence_cutoff_threshold] = np.NaN

    number_of_nans = np.sum(np.isnan(mediapipe_2d_data))
    number_of_points = np.prod(mediapipe_2d_data.shape)
    percentage_that_are_nans = (np.sum(np.isnan(mediapipe_2d_data)) / number_of_points) * 100
    logger.info(
        f"After thresholding `mediapipe_2d` with a confidence threshold {mediapipe_confidence_cutoff_threshold}, it has {number_of_nans} NaN values out of {number_of_points} ({percentage_that_are_nans} %)"
    )
    return mediapipe_2d_data


def remove_3d_data_with_high_reprojection_error(
    data3d_numFrames_numTrackedPoints_XYZ: np.ndarray,
    data3d_numFrames_numTrackedPoints_reprojectionError: np.ndarray,
):
    logger.info("Removing 3D data with high reprojection error")
    mean_reprojection_error_per_frame = np.nanmean(
        data3d_numFrames_numTrackedPoints_reprojectionError,
        axis=1,
    )

    reprojection_error_mean = np.nanmean(mean_reprojection_error_per_frame)
    reprojection_error_median = np.nanmedian(mean_reprojection_error_per_frame)
    reprojection_error_std = np.nanstd(mean_reprojection_error_per_frame)

    median_absolute_deviation = np.nanmedian(np.abs(mean_reprojection_error_per_frame - reprojection_error_median))

    logger.info(
        f"\nInitial reprojection error - \nmean: {reprojection_error_mean:.3f},\nstandard deviation: {reprojection_error_std:.3f},\nmedian: {reprojection_error_median}\nmedian absolute deviation: {median_absolute_deviation:.3f}"
    )

    error_threshold = reprojection_error_median + 3 * median_absolute_deviation

    number_of_nans_before_thresholding = np.sum(np.isnan(data3d_numFrames_numTrackedPoints_XYZ))

    # replace points with high reprojection error with `np.nan`
    data3d_numFrames_numTrackedPoints_XYZ[
        data3d_numFrames_numTrackedPoints_reprojectionError > error_threshold
    ] = np.nan

    number_of_nans_after_thresholding = np.sum(np.isnan(data3d_numFrames_numTrackedPoints_XYZ))
    percentage_of_nans_removed = (
        (number_of_nans_before_thresholding - number_of_nans_after_thresholding)
        / number_of_nans_before_thresholding
        * 100
    )

    logger.info(f"Removing points with reprojection error > {error_threshold:.3f}")
    logger.info(f"Number of NaNs before thresholding: {number_of_nans_before_thresholding}")
    logger.info(
        f"Number of NaNs after thresholding: {number_of_nans_after_thresholding} ({percentage_of_nans_removed:.2f} %)"
    )

    return data3d_numFrames_numTrackedPoints_XYZ


def triangulate_3d_data(
    anipose_calibration_object,
    mediapipe_2d_data: np.ndarray,
    output_data_folder_path: Union[str, Path],
    mediapipe_confidence_cutoff_threshold: float,
    use_triangulate_ransac: bool = False,
    kill_event : multiprocessing.Event = None,
):

    number_of_cameras = mediapipe_2d_data.shape[0]
    number_of_frames = mediapipe_2d_data.shape[1]
    number_of_tracked_points = mediapipe_2d_data.shape[2]
    number_of_spatial_dimensions = mediapipe_2d_data.shape[3]

    if not number_of_spatial_dimensions == 2:
        logger.error(
            f"This is supposed to be 2D data but, number_of_spatial_dimensions: {number_of_spatial_dimensions}"
        )
        raise Exception

    mediapipe_2d_data = threshold_by_confidence(
        mediapipe_2d_data=mediapipe_2d_data,
        mediapipe_confidence_cutoff_threshold=mediapipe_confidence_cutoff_threshold,
    )

    # reshape data to collapse across 'frames' so it becomes [number_of_cameras,
    # number_of_2d_points(numFrames*numPoints), XY]
    data2d_flat = mediapipe_2d_data.reshape(number_of_cameras, -1, 2)

    logger.info(
        f"Reconstructing 3d points from 2d points with shape: \n"
        f"number_of_cameras: {number_of_cameras},\n"
        f"number_of_frames: {number_of_frames}, \n"
        f"number_of_tracked_points: {number_of_tracked_points},\n"
        f"number_of_spatial_dimensions: {number_of_spatial_dimensions}"
    )

    if use_triangulate_ransac:
        logger.info("Using `triangulate_ransac` method")
        data3d_flat = anipose_calibration_object.triangulate_ransac(data2d_flat, progress=True, kill_event=kill_event)
    else:
        logger.info("Using simple `triangulate` method ")
        data3d_flat = anipose_calibration_object.triangulate(data2d_flat, progress=True, kill_event=kill_event)

    spatial_data3d_numFrames_numTrackedPoints_XYZ_og = data3d_flat.reshape(
        number_of_frames, number_of_tracked_points, 3
    )

    data3d_reprojectionError_flat = anipose_calibration_object.reprojection_error(data3d_flat, data2d_flat, mean=True)

    reprojection_error_data3d_numFrames_numTrackedPoints = data3d_reprojectionError_flat.reshape(
        number_of_frames, number_of_tracked_points
    )

    spatial_data3d_numFrames_numTrackedPoints_XYZ = remove_3d_data_with_high_reprojection_error(
        data3d_numFrames_numTrackedPoints_XYZ=spatial_data3d_numFrames_numTrackedPoints_XYZ_og,
        data3d_numFrames_numTrackedPoints_reprojectionError=reprojection_error_data3d_numFrames_numTrackedPoints,
    )

    save_mediapipe_3d_data_to_npy(
        data3d_numFrames_numTrackedPoints_XYZ=spatial_data3d_numFrames_numTrackedPoints_XYZ,
        data3d_numFrames_numTrackedPoints_reprojectionError=reprojection_error_data3d_numFrames_numTrackedPoints,
        path_to_folder_where_data_will_be_saved=output_data_folder_path,
    )
    # TODO - don't output multiple variables, make this a dataclass/BAG pattern or something
    return (
        spatial_data3d_numFrames_numTrackedPoints_XYZ,
        reprojection_error_data3d_numFrames_numTrackedPoints,
    )


def save_mediapipe_3d_data_to_npy(
    data3d_numFrames_numTrackedPoints_XYZ: np.ndarray,
    data3d_numFrames_numTrackedPoints_reprojectionError: np.ndarray,
    path_to_folder_where_data_will_be_saved: Union[str, Path],
):
    path_to_folder_where_data_will_be_saved = Path(path_to_folder_where_data_will_be_saved)
    Path(path_to_folder_where_data_will_be_saved).mkdir(parents=True, exist_ok=True)  # save spatial XYZ data
    mediapipe_3dData_save_path = path_to_folder_where_data_will_be_saved / RAW_MEDIAPIPE_3D_NPY_FILE_NAME

    logger.info(f"saving: {mediapipe_3dData_save_path}")
    np.save(str(mediapipe_3dData_save_path), data3d_numFrames_numTrackedPoints_XYZ)

    # save reprojection error
    mediapipe_reprojection_error_save_path = (
        path_to_folder_where_data_will_be_saved / MEDIAPIPE_REPROJECTION_ERROR_NPY_FILE_NAME
    )

    logger.info(f"saving: {mediapipe_reprojection_error_save_path}")
    np.save(
        str(mediapipe_reprojection_error_save_path),
        data3d_numFrames_numTrackedPoints_reprojectionError,
    )


if __name__ == "__main__":
    from freemocap.core_processes.capture_volume_calibration.anipose_camera_calibration.get_anipose_calibration_object import (
        load_anipose_calibration_toml_from_path,
    )
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--mediapipe_2d_data_path",
        type=str,
        help="path to mediapipe 2d data",
        required=True,
    )
    parser.add_argument(
        "--output_data_folder_path",
        type=str,
        help="path to output folder",
        required=False,
    )
    parser.add_argument(
        "--calibration_file_path",
        type=str,
        help="path to calibration file",
        required=False,
    )
    parser.add_argument(
        "--mediapipe_confidence_cutoff_threshold",
        type=float,
        help="confidence cutoff threshold",
        required=False,
    )
    parser.add_argument(
        "--save_data_as_csv",
        type=bool,
        help="save data as csv",
        required=False,
    )
    parser.add_argument(
        "--use_triangulate_ransac",
        type=bool,
        help="use triangulate ransac anipose method ",
        required=False,
    )
    args = parser.parse_args()

    mediapipe_2d_data = np.load(args.mediapipe_2d_data_path)

    if args.mediapipe_confidence_cutoff_threshold is None:
        args.mediapipe_confidence_cutoff_threshold = 0.7  # default

    if args.output_data_folder_path is None:
        args.output_data_folder_path = Path(args.mediapipe_2d_data_path).parent

    if args.save_data_as_csv is None:
        args.save_data_as_csv = True  # default

    if args.use_triangulate_ransac is None:
        args.use_triangulate_ransac = True

    if args.calibration_file_path:
        anipose_calibration_object = load_anipose_calibration_toml_from_path(args.calibration_file_path)
    else:
        anipose_calibration_object = load_anipose_calibration_toml_from_path(
            Path(args.mediapipe_2d_data_path).parent.parent / "camera_calibration_data.toml"
        )

    triangulate_3d_data(
        anipose_calibration_object=anipose_calibration_object,
        mediapipe_2d_data=mediapipe_2d_data,
        output_data_folder_path=args.output_data_folder_path,
        mediapipe_confidence_cutoff_threshold=args.mediapipe_confidence_cutoff_threshold,
        use_triangulate_ransac=args.use_triangulate_ransac,
    )



import logging
import multiprocessing

from PyQt6.QtCore import pyqtSignal, QThread

from freemocap.core_processes.process_motion_capture_videos.process_recording_folder import (
    process_recording_folder,
)
from freemocap.parameter_info_models.recording_processing_parameter_models import RecordingProcessingParameterModel

logger = logging.getLogger(__name__)


class ProcessMotionCaptureDataThreadWorker(QThread):
    finished = pyqtSignal()
    in_progress = pyqtSignal(str)

    def __init__(self, session_processing_parameters: RecordingProcessingParameterModel,
                 kill_event: multiprocessing.Event, parent=None):
        super().__init__()
        self._session_processing_parameters = session_processing_parameters
        self._kill_event = kill_event


    @property
    def work_done(self):
        return self._work_done

    def _emit_in_progress_data(self, message: str):
        self.in_progress.emit(message)

    def run(self):
        logger.info(
            f"Beginning processing of motion capture data with parameters: {self._session_processing_parameters.__dict__}"
        )
        self._kill_event.clear()

        process_recording_folder(
            recording_processing_parameter_model=self._session_processing_parameters,
            kill_event=self._kill_event,
        )

        logger.info("Finished processing session folder!")

        self.finished.emit()
        self._work_done = True

    def kill(self):
        self.terminate()



import logging
import threading
from pathlib import Path
from typing import Union

from PyQt6.QtCore import pyqtSignal, QThread

from freemocap.core_processes.capture_volume_calibration.charuco_stuff.charuco_board_definition import (
    CharucoBoardDefinition,
)
from freemocap.core_processes.capture_volume_calibration.run_anipose_capture_volume_calibration import (
    run_anipose_capture_volume_calibration,
)

logger = logging.getLogger(__name__)


class AniposeCalibrationThreadWorker(QThread):
    finished = pyqtSignal()
    in_progress = pyqtSignal(str)

    def __init__(
        self,
        calibration_videos_folder_path: Union[str, Path],
        charuco_square_size: Union[int, float],
        kill_thread_event: threading.Event,
        charuco_board_definition: CharucoBoardDefinition = CharucoBoardDefinition(),
    ):
        super().__init__()
        logger.info(
            f"Initializing Anipose Calibration Thread Worker for videos in path {calibration_videos_folder_path}"
        )
        self._kill_thread_event = kill_thread_event
        self._charuco_board_definition = charuco_board_definition
        self._charuco_square_size = charuco_square_size
        self._calibration_videos_folder_path = calibration_videos_folder_path

        self._work_done = False

    @property
    def work_done(self):
        return self._work_done

    def _emit_in_progress_data(self, message: str):
        self.in_progress.emit(message)

    def run(self):
        logger.info("Beginning Anipose calibration with Charuco Square Size (mm): {}".format(self._charuco_square_size))

        try:


            run_anipose_capture_volume_calibration(
                charuco_board_definition=self._charuco_board_definition,
                charuco_square_size=self._charuco_square_size,
                calibration_videos_folder_path=self._calibration_videos_folder_path,
                pin_camera_0_to_origin=True,
                progress_callback=self._emit_in_progress_data,
            )



        except Exception as e:
            logger.exception("something went wrong in the anipose calibration")
            logger.exception(e)

        self.finished.emit()
        self._work_done = True

        logger.info("Anipose Calibration Complete")




from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import (
    QDockWidget,
    QVBoxLayout,
    QWidget,
    QTabWidget,
)
from skellycam import SkellyCamParameterTreeWidget

from freemocap.gui.qt.widgets.control_panel.process_mocap_data_panel.process_motion_capture_data_panel import (
    ProcessMotionCaptureDataPanel,
)
from freemocap.system.paths_and_files_names import (
    HAMMER_AND_WRENCH_EMOJI_STRING,
    CAMERA_WITH_FLASH_EMOJI_STRING,
    THREE_HEARTS_EMOJI_STRING,
    ROBOT_EMOJI_STRING,
    WIND_EMOJI_STRING,
    SPARKLES_EMOJI_STRING,
)


class ControlPanelDockWidget(QDockWidget):
    def __init__(self, **kwargs):
        super().__init__("Control", parent=kwargs.get("parent"))

        self.setFeatures(QDockWidget.DockWidgetFeature.DockWidgetMovable)
        self.setFeatures(QDockWidget.DockWidgetFeature.DockWidgetFloatable)

        self.setAllowedAreas(Qt.DockWidgetArea.LeftDockWidgetArea | Qt.DockWidgetArea.RightDockWidgetArea)

        self._control_panel_widget = ControlPanelWidget(**kwargs)
        self.setWidget(self._control_panel_widget)

    @property
    def tool_box(self):
        return self._control_panel_widget.tab_widget


class ControlPanelWidget(QWidget):
    def __init__(
        self,
        camera_configuration_parameter_tree_widget: SkellyCamParameterTreeWidget,
        # calibration_control_panel: CalibrationControlPanel,
        process_motion_capture_data_panel: ProcessMotionCaptureDataPanel,
        visualize_data_widget: QWidget,
        parent=None,
    ):
        super().__init__(parent=parent)

        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self._tab_widget = QTabWidget(parent=self)
        self._layout.addWidget(self._tab_widget)

        self._tab_widget.addTab(
            camera_configuration_parameter_tree_widget,
            f"Camera Configuration",
        )
        # self._tool_box.addTab(calibration_control_panel, "Capture Volume Calibration")
        self._tab_widget.addTab(
            process_motion_capture_data_panel, f"Process Data"
        )
        self._tab_widget.addTab(visualize_data_widget, f"Export Data")

        self._tab_widget.setProperty("control_panel_tabs", True)
        self.style().polish(self._tab_widget)

    @property
    def tab_widget(self):
        return self._tab_widget



import logging
import multiprocessing
import shutil
from pathlib import Path
from typing import Union, List

from PyQt6.QtCore import Qt, pyqtSlot
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import (
    QDockWidget,
    QMainWindow,
    QFileDialog,
    QTabWidget, QWidget, QHBoxLayout, )
from skelly_viewer import SkellyViewer
from skellycam import (
    SkellyCamParameterTreeWidget,
    SkellyCamWidget,
    SkellyCamControllerWidget,
)
from tqdm import tqdm

from freemocap.core_processes.export_data.blender_stuff.export_to_blender import (
    export_to_blender,
)
from freemocap.core_processes.export_data.blender_stuff.get_best_guess_of_blender_path import \
    get_best_guess_of_blender_path
from freemocap.gui.qt.actions_and_menus.actions import Actions
from freemocap.gui.qt.actions_and_menus.menu_bar import MenuBar
from freemocap.gui.qt.style_sheet.css_file_watcher import CSSFileWatcher
from freemocap.gui.qt.style_sheet.scss_file_watcher import SCSSFileWatcher
from freemocap.gui.qt.style_sheet.set_css_style_sheet import apply_css_style_sheet
from freemocap.gui.qt.utilities.get_qt_app import get_qt_app
from freemocap.gui.qt.utilities.update_most_recent_recording_toml import (
    update_most_recent_recording_toml,
)
from freemocap.gui.qt.widgets.active_recording_widget import ActiveRecordingInfoWidget
from freemocap.gui.qt.widgets.camera_controller_group_box import CameraControllerGroupBox
from freemocap.gui.qt.widgets.central_tab_widget import CentralTabWidget
from freemocap.gui.qt.widgets.control_panel.control_panel_dock_widget import (
    ControlPanelWidget,
)
from freemocap.gui.qt.widgets.control_panel.process_mocap_data_panel.process_motion_capture_data_panel import (
    ProcessMotionCaptureDataPanel,
)
from freemocap.gui.qt.widgets.control_panel.visualization_control_panel import VisualizationControlPanel
from freemocap.gui.qt.widgets.directory_view_widget import DirectoryViewWidget
from freemocap.gui.qt.widgets.import_videos_window import ImportVideosWizard
from freemocap.gui.qt.widgets.log_view_widget import LogViewWidget
from freemocap.gui.qt.widgets.home_widget import (
    HomeWidget,
)
from freemocap.parameter_info_models.recording_info_model import (
    RecordingInfoModel,
)
from freemocap.parameter_info_models.recording_processing_parameter_models import (
    RecordingProcessingParameterModel,
)
from freemocap.system.paths_and_files_names import (
    get_scss_stylesheet_path,
    get_css_stylesheet_path,
    get_most_recent_recording_path,
    PATH_TO_FREEMOCAP_LOGO_SVG,
    get_blender_file_path,
    get_recording_session_folder_path,
    DIRECTORY_EMOJI_STRING,
    GEAR_EMOJI_STRING,
    COOL_EMOJI_STRING,
)
# reboot GUI method based on this - https://stackoverflow.com/a/56563926/14662833
from freemocap.system.start_file import open_file
from freemocap.system.user_data.pipedream_pings import PipedreamPings

EXIT_CODE_REBOOT = -123456789

logger = logging.getLogger(__name__)


class FreemocapMainWindow(QMainWindow):
    def __init__(
            self,
            freemocap_data_folder_path: Union[str, Path],
            pipedream_pings: PipedreamPings,
            parent=None,
    ):

        logger.info("Initializing QtGUIMainWindow")
        super().__init__(parent=parent)

        self.setGeometry(100, 100, 1280, 720)
        self.setWindowIcon(QIcon(PATH_TO_FREEMOCAP_LOGO_SVG))
        self.setWindowTitle("freemocap \U0001F480 \U00002728")

        dummy_widget = QWidget()
        self._layout = QHBoxLayout()
        dummy_widget.setLayout(self._layout)
        self.setCentralWidget(dummy_widget)

        self._scss_file_watcher = self._set_up_stylesheet()

        self._freemocap_data_folder_path = freemocap_data_folder_path
        self._pipedream_pings = pipedream_pings

        self._kill_thread_event = multiprocessing.Event()

        self._active_recording_info_widget = ActiveRecordingInfoWidget(parent=self)
        self._active_recording_info_widget.new_active_recording_selected_signal.connect(
            self._handle_new_active_recording_selected
        )
        self._directory_view_widget = self._create_directory_view_widget()
        self._directory_view_widget.expand_directory_to_path(get_recording_session_folder_path())
        self._directory_view_widget.new_active_recording_selected_signal.connect(
            self._active_recording_info_widget.set_active_recording
        )

        self._actions = Actions(freemocap_main_window=self)

        self._menu_bar = MenuBar(actions=self._actions, parent=self)
        self.setMenuBar(self._menu_bar)

        self.statusBar().showMessage(
            "Watch the terminal output for status updates, we're working on integrating better status updates into the GUI")

        self._central_tab_widget = self._create_central_tab_widget()
        self.setCentralWidget(self._central_tab_widget)

        self._tools_dock_widget = self._create_tools_dock_widget()
        self.addDockWidget(Qt.DockWidgetArea.RightDockWidgetArea, self._tools_dock_widget)
        # self._tools_dock_tab_widget = QTabWidget(self)

        self._control_panel_widget = self._create_control_panel_widget()
        self._tools_dock_widget.setWidget(self._control_panel_widget)
        # self._tools_dock_tab_widget.addTab(self._control_panel_widget, f"Control Panel")
        #
        # self._tools_dock_tab_widget.addTab(self._directory_view_widget, f"Directory View")
        # self._tools_dock_tab_widget.addTab(
        #     self._active_recording_info_widget, f"Active Recording Info"
        # )

        # self.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, self._directory_view_dock_widget)

        # self._jupyter_console_widget = JupyterConsoleWidget(parent=self)
        # jupyter_console_dock_widget = QDockWidget("Jupyter IPython Console", self)
        # self.addDockWidget(
        #     Qt.DockWidgetArea.BottomDockWidgetArea, jupyter_console_dock_widget
        # )
        # jupyter_console_dock_widget.setWidget(self._jupyter_console_widget)

        self._log_view_widget = LogViewWidget(parent=self)
        log_view_dock_widget = QDockWidget("Log View", self)
        self.addDockWidget(Qt.DockWidgetArea.BottomDockWidgetArea, log_view_dock_widget)
        log_view_dock_widget.setWidget(self._log_view_widget)
        log_view_dock_widget.setFeatures(
            QDockWidget.DockWidgetFeature.DockWidgetMovable | QDockWidget.DockWidgetFeature.DockWidgetFloatable
        )

    def _create_tools_dock_widget(self):
        tools_dock_widget = QDockWidget("Control Panel", self)
        tools_dock_widget.setAllowedAreas(Qt.DockWidgetArea.LeftDockWidgetArea | Qt.DockWidgetArea.RightDockWidgetArea)
        tools_dock_widget.setFeatures(
            QDockWidget.DockWidgetFeature.DockWidgetMovable | QDockWidget.DockWidgetFeature.DockWidgetFloatable
        )
        return tools_dock_widget

    def _handle_videos_saved_to_this_folder_signal(self, folder_path: str):
        logger.info(f"Videos saved to this folder signal received: {folder_path}")

        self._active_recording_info_widget.set_active_recording(recording_folder_path=folder_path)

        if (
                self._controller_group_box.auto_process_videos_checked
                and self._controller_group_box.mocap_videos_radio_button_checked
        ):
            logger.info("'Auto process videos' checkbox is checked - triggering 'Process Motion Capture Data' button")
            self._process_motion_capture_data_panel.process_motion_capture_data_button.click()
        elif self._controller_group_box.calibration_videos_radio_button_checked:
            logger.info("Processing calibration videos")
            self._process_motion_capture_data_panel.calibrate_from_active_recording(
                charuco_square_size_mm=float(self._controller_group_box.charuco_square_size)
            )

    def _handle_processing_finished_signal(self):
        logger.info("Processing finished")
        if self._controller_group_box.auto_open_in_blender_checked and not self._kill_thread_event.is_set():
            logger.info("'Auto process videos' checkbox is checked - triggering 'Create Blender Scene'")
            self._export_active_recording_to_blender()

    def handle_start_new_session_action(self):
        # self._central_tab_widget.set_welcome_tab_enabled(True)
        self._central_tab_widget.set_camera_view_tab_enabled(True)
        self._central_tab_widget.setCurrentIndex(1)
        self._controller_group_box.show()
        self._skellycam_widget.detect_available_cameras()

    def update(self):
        super().update()

        try:
            if not self._skellycam_widget.is_recording:
                self._controller_group_box.update_recording_name_string()
        except Exception as e:
            logger.exception(e)

    def _set_up_stylesheet(self):
        apply_css_style_sheet(self, get_css_stylesheet_path())
        scss_file_watcher = SCSSFileWatcher(path_to_scss_file=get_scss_stylesheet_path(), path_to_css_file=get_css_stylesheet_path(), parent=self)
        return scss_file_watcher

    def _create_new_synchronized_videos_folder(self) -> str:
        new_recording_folder_path = self._controller_group_box.get_new_recording_path()
        logger.info(f"Creating new recording folder at: {new_recording_folder_path}")
        self._active_recording_info_widget.set_active_recording(recording_folder_path=new_recording_folder_path)
        return self._active_recording_info_widget.active_recording_info.synchronized_videos_folder_path

    def _create_central_tab_widget(self):

        self._home_widget = HomeWidget(actions=self._actions, parent=self)

        self._skellycam_widget = SkellyCamWidget(
            self._create_new_synchronized_videos_folder,
            parent=self,
        )
        self._skellycam_widget.videos_saved_to_this_folder_signal.connect(
            self._handle_videos_saved_to_this_folder_signal
        )

        self._skellycam_controller_widget = SkellyCamControllerWidget(
            self._skellycam_widget,
            parent=self,
        )

        self._controller_group_box = CameraControllerGroupBox(
            skellycam_controller=self._skellycam_controller_widget, parent=self
        )

        self._skelly_viewer_widget = SkellyViewer()

        center_tab_widget = CentralTabWidget(
            parent=self,
            skelly_cam_widget=self._skellycam_widget,
            camera_controller_widget=self._controller_group_box,
            welcome_to_freemocap_widget=self._home_widget,
            skelly_viewer_widget=self._skelly_viewer_widget,
            directory_view_widget=self._directory_view_widget,
            active_recording_info_widget=self._active_recording_info_widget
        )

        center_tab_widget.set_welcome_tab_enabled(True)
        center_tab_widget.set_camera_view_tab_enabled(True)
        center_tab_widget.set_visualize_data_tab_enabled(True)

        return center_tab_widget

    def _create_directory_view_widget(self):
        return DirectoryViewWidget(
            top_level_folder_path=self._freemocap_data_folder_path,
            get_active_recording_info_callable=self._active_recording_info_widget.get_active_recording_info,
        )

    # def _create_tool_bar(self):
    #     self._calibration_control_panel = CalibrationControlPanel(
    #         get_active_recording_info_callable=self._active_recording_info_widget.get_active_recording_info,
    #     )
    #     self._process_motion_capture_data_panel = ProcessMotionCaptureDataPanel(
    #         recording_processing_parameters=RecordingProcessingParameterModel(),
    #         get_active_recording_info=self._active_recording_info_widget.get_active_recording_info,
    #     )
    #     self._process_motion_capture_data_panel.processing_finished_signal.connect(
    #         self._handle_processing_finished_signal
    #     )
    #     return ToolBar(calibration_control_panel=self._calibration_control_panel,
    #                    process_motion_capture_data_panel=self._process_motion_capture_data_panel,
    #                    visualize_data_widget=self._create_visualization_control_panel(),
    #                    directory_view_widget=self._create_directory_view_widget(),
    #                    parent=self)

    def _create_control_panel_widget(self):

        self._camera_configuration_parameter_tree_widget = SkellyCamParameterTreeWidget(self._skellycam_widget)

        # self._calibration_control_panel = CalibrationControlPanel(
        #     get_active_recording_info_callable=self._active_recording_info_widget.get_active_recording_info,
        # )

        self._process_motion_capture_data_panel = ProcessMotionCaptureDataPanel(
            recording_processing_parameters=RecordingProcessingParameterModel(),
            get_active_recording_info=self._active_recording_info_widget.get_active_recording_info,
            kill_thread_event=self._kill_thread_event,
        )
        self._process_motion_capture_data_panel.processing_finished_signal.connect(
            self._handle_processing_finished_signal
        )

        self._visualization_control_panel = VisualizationControlPanel(parent=self,
                                                                      blender_executable_path=get_best_guess_of_blender_path())
        self._visualization_control_panel.export_to_blender_button.clicked.connect(
            self._export_active_recording_to_blender
        )

        return ControlPanelWidget(
            camera_configuration_parameter_tree_widget=self._camera_configuration_parameter_tree_widget,
            # calibration_control_panel=self._calibration_control_panel,
            process_motion_capture_data_panel=self._process_motion_capture_data_panel,
            visualize_data_widget=self._visualization_control_panel,
            parent=self,
        )

    def _export_active_recording_to_blender(self):
        logger.info("Exporting active recording to Blender...")
        recording_path = self._active_recording_info_widget.get_active_recording_info(return_path=True)

        if self._visualization_control_panel.blender_executable_path is None:
            logger.error("Blender executable path is None!")
            return

        self._visualization_control_panel.get_user_selected_method_string()
        export_to_blender(
            recording_folder_path=recording_path,
            blender_file_path=get_blender_file_path(recording_path),
            blender_exe_path=self._visualization_control_panel.blender_executable_path,
            method=self._visualization_control_panel.get_user_selected_method_string(),
        )

        if (
                self._controller_group_box.auto_open_in_blender_checked
                # or self._visualization_control_panel.open_in_blender_automatically_box_is_checked
        ):
            if Path(self._active_recording_info_widget.active_recording_info.blender_file_path).exists():
                open_file(self._active_recording_info_widget.active_recording_info.blender_file_path)
            else:
                logger.error("Blender file does not exist! Did something go wrong in the `export_to_blender` call above?")

    def _handle_new_active_recording_selected(self, recording_info_model: RecordingInfoModel):
        logger.info(f"New active recording selected: {recording_info_model.path}")

        # self._calibration_control_panel.update_calibrate_from_active_recording_button_text()
        self._pipedream_pings.update_pings_dict(key=recording_info_model.name,
                                                value=self._active_recording_info_widget.active_recording_info.status_check)

        path = Path(recording_info_model.path)
        path.mkdir(parents=True, exist_ok=True)

        if str(path.parent) != get_recording_session_folder_path():
            self._directory_view_widget.set_folder_as_root(path.parent)
        else:
            self._directory_view_widget.set_folder_as_root(path)

        if Path(recording_info_model.synchronized_videos_folder_path).exists():
            self._directory_view_widget.expand_directory_to_path(recording_info_model.synchronized_videos_folder_path)
        else:
            self._directory_view_widget.expand_directory_to_path(recording_info_model.path)

        self._active_recording_info_widget.update_parameter_tree()
        # self._recording_name_label.setText(f"Recording Name: {recording_info_model.name}")
        self._update_skelly_viewer_widget()
        self._directory_view_widget.handle_new_active_recording_selected()

        try:
            self._process_motion_capture_data_panel.update_calibration_path()
        except Exception as e:
            logger.error(e)



        update_most_recent_recording_toml(recording_info_model=recording_info_model)

    def _update_skelly_viewer_widget(self):
        active_recording_info = self._active_recording_info_widget.active_recording_info


        if active_recording_info.data3d_status_check:
            self._skelly_viewer_widget.load_skeleton_data(
                mediapipe_skeleton_npy_path=active_recording_info.mediapipe_3d_data_npy_file_path
            )


        if active_recording_info.data2d_status_check:
            self._skelly_viewer_widget.generate_video_display(
                video_folder_path=active_recording_info.annotated_videos_folder_path
            )


        elif active_recording_info.synchronized_videos_status_check:
            self._skelly_viewer_widget.generate_video_display(
                video_folder_path=active_recording_info.synchronized_videos_folder_path
            )


    def reboot_gui(self):
        logger.info("Rebooting GUI... ")
        get_qt_app().exit(EXIT_CODE_REBOOT)

    def kill_running_threads_and_processes(self):
        logger.info("Killing running threads and processes... ")
        try:
            self._skellycam_widget.close()
        except Exception as e:
            logger.error(f"Error killing running threads and processes: {e}")

        self._kill_thread_event.set()

    def handle_load_most_recent_recording(self):
        logger.info("`Load Most Recent Recording` QAction triggered")
        most_recent_recording_path = get_most_recent_recording_path()

        if most_recent_recording_path is None:
            logger.error(f"`get_most_recent_recording_path()` return `None`!")
            return

        self._active_recording_info_widget.set_active_recording(recording_folder_path=get_most_recent_recording_path())
        self._central_tab_widget.setCurrentIndex(2)
        self._control_panel_widget.tab_widget.setCurrentWidget(self._process_motion_capture_data_panel)

    def open_load_existing_recording_dialog(self):
        # from this tutorial - https://www.youtube.com/watch?v=gg5TepTc2Jg&t=649s
        logger.info("Opening `Load Existing Recording` dialog... ")
        user_selected_directory = QFileDialog.getExistingDirectory(
            self,
            "Select a recording folder",
            str(get_recording_session_folder_path()),
        )
        if len(user_selected_directory) == 0:
            logger.info("User cancelled `Load Existing Recording` dialog")
            return

        logger.info(f"User selected recording path:{user_selected_directory}")

        self._active_recording_info_widget.set_active_recording(recording_folder_path=user_selected_directory)
        self._central_tab_widget.setCurrentIndex(2)

    def open_import_videos_dialog(self):
        # from this tutorial - https://www.youtube.com/watch?v=gg5TepTc2Jg&t=649s
        logger.info("Opening `Import Videos` dialog... ")

        import_videos_path = QFileDialog.getExistingDirectory(
            self,
            "Select a folder containing synchronized videos (each video must have *exactly* the same number of frames)",
            str(Path.home()),
        )

        if len(import_videos_path) == 0:
            logger.info("User cancelled `Import Videos` dialog")
            return

        self._import_videos_window = ImportVideosWizard(parent=self,
                                                        import_videos_path=import_videos_path,
                                                        )
        self._import_videos_window.folder_to_save_videos_to_selected.connect(self._handle_import_videos)
        text = self._import_videos_window.exec()


    @pyqtSlot(list, str)
    def _handle_import_videos(self, video_paths: List[str], folder_to_save_videos: str):

        folder_to_save_videos = Path(folder_to_save_videos)
        folder_to_save_videos.mkdir(parents=True, exist_ok=True)

        if len(video_paths) == 0:
            logger.error("No videos to import!")
            return

        for video_path in tqdm(video_paths,
                               desc="Importing videos...",
                               colour=[255, 128, 0],
                               unit="video",
                               unit_scale=True,
                               leave=False
                               ):

            if not Path(video_path).exists():
                logger.error(f"{video_path} does not exist!")
                return

            destination_path = folder_to_save_videos / Path(video_path).name
            logger.info(f"Copying video from {video_path} to {destination_path}")

            shutil.copy(video_path, destination_path)

        self._active_recording_info_widget.set_active_recording(
            recording_folder_path=Path(folder_to_save_videos).parent)

    def closeEvent(self, a0) -> None:
        logger.info("Main window `closeEvent` detected")

        self._pipedream_pings.update_pings_dict(key="gui_closed", value=True)
        self._pipedream_pings.update_pings_dict(key="active recording status on close",
                                                value=self._active_recording_info_widget.active_recording_info.status_check)
        self._pipedream_pings.send_pipedream_ping()

        remove_empty_directories(get_recording_session_folder_path())

        try:
            self._skellycam_widget.close()
        except Exception as e:
            logger.error(f"Error while closing the viewer widget: {e}")
        super().closeEvent(a0)


def remove_empty_directories(root_dir: Union[str, Path]):
    """
    Recursively remove empty directories from the root directory
    :param root_dir: The root directory to start removing empty directories from
    """
    # logger.debug(f"Searching for empty directories in: {root_dir}")
    for path in Path(root_dir).rglob("*"):
        if path.is_dir() and not any(path.iterdir()):
            logger.info(f"Removing empty directory: {path}")
            path.rmdir()
        elif path.is_dir() and any(path.iterdir()):
            remove_empty_directories(path)
        else:
            continue


if __name__ == "__main__":
    import sys

    from PyQt6.QtWidgets import QApplication

    app = QApplication(sys.argv)
    main_window = FreemocapMainWindow(pipedream_pings=PipedreamPings())
    main_window.show()
    app.exec()
    for process in multiprocessing.active_children():
        logger.info(f"Terminating process: {process}")
        process.terminate()
    sys.exit()
